from __future__ import annotations

import abc
import threading
import weakref
from enum import Enum
from typing import TYPE_CHECKING, Callable

import pyglet
from pyglet import gl
from pyglet.gl import gl_info

if TYPE_CHECKING:
    from _ctypes import Array
    from pyglet.display.base import Canvas
    from pyglet.gl.gl_info import GLInfo
    from pyglet.graphics.shader import ShaderProgram


class OpenGLAPI(Enum):
    """The OpenGL API backend to use."""
    OPENGL = 1
    OPENGL_ES = 2


class Config:
    """Graphics configuration.

    A Config stores the preferences for OpenGL attributes such as the
    number of auxiliary buffers, size of the colour and depth buffers,
    double buffering, stencilling, multi- and super-sampling, and so on.

    Different platforms support a different set of attributes, so these
    are set with a string key and a value which is integer or boolean.
    """

    #: Specify the presence of a back-buffer for every color buffer.
    double_buffer: bool
    #: Specify the presence of separate left and right buffer sets.
    stereo: bool
    #: Total bits per sample per color buffer.
    buffer_size: int
    #: The number of auxiliary color buffers.
    aux_buffers: int
    #: The number of multisample buffers.
    sample_buffers: int
    #: The number of samples per pixel, or 0 if there are no multisample buffers.
    samples: int
    #: Bits per sample per buffer devoted to the red component.
    red_size: int
    #: Bits per sample per buffer devoted to the green component.
    green_size: int
    #: Bits per sample per buffer devoted to the blue component.
    blue_size: int
    #: Bits per sample per buffer devoted to the alpha component.
    alpha_size: int
    #: Bits per sample in the depth buffer.
    depth_size: int
    #: Bits per sample in the stencil buffer.
    stencil_size: int
    #: Bits per pixel devoted to the red component in the accumulation buffer.
    accum_red_size: int
    #: Bits per pixel devoted to the green component in the accumulation buffer.
    accum_green_size: int
    #: Bits per pixel devoted to the blue component in the accumulation buffer.
    accum_blue_size: int
    #: Bits per pixel devoted to the alpha component in the accumulation buffer.
    accum_alpha_size: int

    _attribute_names = (
        'double_buffer',
        'stereo',
        'buffer_size',
        'aux_buffers',
        'sample_buffers',
        'samples',
        'red_size',
        'green_size',
        'blue_size',
        'alpha_size',
        'depth_size',
        'stencil_size',
        'accum_red_size',
        'accum_green_size',
        'accum_blue_size',
        'accum_alpha_size',
        'major_version',
        'minor_version',
        'forward_compatible',
        'opengl_api',
        'debug',
        'transparent_framebuffer',
    )

    #: The OpenGL major version.
    major_version: int
    #: The OpenGL minor version.
    minor_version: int
    #: Whether to use forward compatibility mode.
    forward_compatible: bool
    #: The OpenGL API, such as "gl" or "gles".
    opengl_api: str
    #: Debug mode.
    debug: bool
    #: If the framebuffer should be transparent.
    transparent_framebuffer: bool

    def __init__(self, **kwargs: float) -> None:
        """Create a template config with the given attributes.

        Specify attributes as keyword arguments, for example::

            template = Config(double_buffer=True)

        """
        for name in self._attribute_names:
            if name in kwargs:
                setattr(self, name, kwargs[name])
            else:
                setattr(self, name, None)

        self.opengl_api = self.opengl_api or "gl"

    def get_gl_attributes(self) -> list[tuple[str, bool | int | str]]:
        """Return a list of attributes set on this config.

        The attributes are returned as a list of tuples, containing
        the name and values. Any unset attributes will have a value
        of ``None``.
        """
        return [(name, getattr(self, name)) for name in self._attribute_names]

    @abc.abstractmethod
    def match(self, canvas: Canvas) -> list[DisplayConfig]:
        """Return a list of matching complete configs for the given canvas."""

    def create_context(self, share: Context | None) -> Context:  # noqa: ARG002
        """Create a GL context that satisfies this configuration.

        Args:
            share:
                If not ``None``, a context with which to share objects with.

        :deprecated: Use `DisplayConfig.create_context`.
        """
        msg = (
            'This config cannot be used to create contexts. '
            'Use Config.match to created a DisplayConfig'
        )
        raise gl.ConfigException(msg)

    def is_complete(self) -> bool:
        """Determine if this config is complete and able to create a context.

        Configs created directly are not complete, they can only serve
        as templates for retrieving a supported config from the system.
        For example, `pyglet.window.Screen.get_matching_configs` returns
        complete configs.

        :deprecated: Use ``isinstance(config, DisplayConfig)``.
        """
        return isinstance(self, DisplayConfig)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.get_gl_attributes()})"


class DisplayConfig(Config, abc.ABC):
    """An OpenGL configuration for a particular display.

    Use ``Config.match`` to obtain an instance of this class.

    .. versionadded:: 1.2
    """
    canvas: Canvas

    def __init__(self, canvas: Canvas, base_config: Config) -> None:
        #: The canvas this config is valid on.
        self.canvas: Canvas = canvas

        self.major_version = base_config.major_version
        self.minor_version = base_config.minor_version
        self.forward_compatible = base_config.forward_compatible
        self.opengl_api = base_config.opengl_api or self.opengl_api
        self.debug = base_config.debug
        self.transparent_framebuffer = base_config.transparent_framebuffer

    @abc.abstractmethod
    def compatible(self, canvas: Canvas) -> bool:
        """Determine compatibility with the canvas."""

    @abc.abstractmethod
    def create_context(self, share: Context) -> Context:
        """Create a GL context that satisfies this configuration.

        Args:
            share:
                If not ``None``, a Context with which to share objects with.
        """

    def is_complete(self) -> bool:
        return True


class ObjectSpace:
    """A container to store shared objects that are to be removed."""

    def __init__(self) -> None:
        """Initialize the context object space."""
        # Objects scheduled for deletion the next time this object space is active.
        self.doomed_textures = []
        self.doomed_buffers = []
        self.doomed_shader_programs = []
        self.doomed_shaders = []
        self.doomed_renderbuffers = []


class Context:
    """A base OpenGL context for drawing.

    Use ``DisplayConfig.create_context`` to create a context.
    """
    #: gl_info.GLInfo instance, filled in on first set_current
    _info: GLInfo | None = None

    #: A container which is shared between all contexts that share GL objects.
    object_space: ObjectSpace
    config: DisplayConfig
    context_share: Context | None

    def __init__(self, config: DisplayConfig, context_share: Context | None = None) -> None:
        """Initialize a context.

        This should only be created through the ``DisplayConfig.create_context`` method.

        Args:
            config:
                An operating system specific config.
            context_share:
                A context to share objects with. Use ``None`` to disable sharing.
        """
        self.config = config
        self.context_share = context_share
        self.canvas = None

        self.doomed_vaos = []
        self.doomed_framebuffers = []

        if context_share:
            self.object_space = context_share.object_space
        else:
            self.object_space = ObjectSpace()

        self._cached_programs = weakref.WeakValueDictionary()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={id(self)}, share={self.context_share})"

    def __enter__(self) -> None:
        self.set_current()

    def __exit__(self, *_args) -> None:  # noqa: ANN002
        return

    def attach(self, canvas: Canvas) -> None:
        if self.canvas is not None:
            self.detach()
        if not self.config.compatible(canvas):
            msg = f'Cannot attach {canvas} to {self}'
            raise RuntimeError(msg)
        self.canvas = canvas

    def detach(self) -> None:
        self.canvas = None

    def set_current(self) -> None:
        """Make this the active Context.

        Setting the Context current will also delete any OpenGL
        objects that have been queued for deletion. IE: any objects
        that were created in this Context, but have been called for
        deletion while another Context was active.
        """
        if not self.canvas:
            msg = 'Canvas has not been attached'
            raise RuntimeError(msg)

        # Not per-thread
        gl.current_context = self

        # Set active context.
        gl_info.set_active_context()

        if not self._info:
            self._info = gl_info.GLInfo()
            self._info.set_active_context()

        if self.object_space.doomed_textures:
            self._delete_objects(self.object_space.doomed_textures, gl.glDeleteTextures)
        if self.object_space.doomed_buffers:
            self._delete_objects(self.object_space.doomed_buffers, gl.glDeleteBuffers)
        if self.object_space.doomed_shader_programs:
            self._delete_objects_one_by_one(self.object_space.doomed_shader_programs,
                                            gl.glDeleteProgram)
        if self.object_space.doomed_shaders:
            self._delete_objects_one_by_one(self.object_space.doomed_shaders, gl.glDeleteShader)
        if self.object_space.doomed_renderbuffers:
            self._delete_objects(self.object_space.doomed_renderbuffers, gl.glDeleteRenderbuffers)

        if self.doomed_vaos:
            self._delete_objects(self.doomed_vaos, gl.glDeleteVertexArrays)
        if self.doomed_framebuffers:
            self._delete_objects(self.doomed_framebuffers, gl.glDeleteFramebuffers)

    # For the static functions below:
    # The garbage collector introduces a race condition.
    # The provided list might be appended to (and only appended to) while this
    # method runs, as it's a `doomed_*` list either on the context or its object
    # space. If `count` wasn't stored in a local, this method might leak objects.
    @staticmethod
    def _delete_objects(list_: list, deletion_func: Callable[[int, Array[gl.GLuint]], None]) -> None:
        """Release all OpenGL objects in the given list.

        Uses the supplied deletion function with the signature ``(GLuint count, GLuint *names)``.
        """
        count = len(list_)
        to_delete = list_[:count]
        del list_[:count]

        deletion_func(count, (gl.GLuint * count)(*to_delete))

    @staticmethod
    def _delete_objects_one_by_one(list_: list, deletion_func: Callable[[gl.GLuint], None]) -> None:
        """Release all OpenGL objects in the given list.

        Similar to ``_delete_objects``, but assumes the deletion function's signature to be ``(GLuint name)``.

        The function is called for each object.
        """
        count = len(list_)
        to_delete = list_[:count]
        del list_[:count]

        for name in to_delete:
            deletion_func(gl.GLuint(name))

    def destroy(self) -> None:
        """Release the Context.

        The context will not be usable after being destroyed.  Each platform
        has its own convention for releasing the context and the buffer(s)
        that depend on it in the correct order; this should never be called
        by an application.
        """
        self.detach()

        if gl.current_context is self:
            gl.current_context = None
            gl_info.remove_active_context()

            # Switch back to shadow context.
            if gl._shadow_window is not None:  # noqa: SLF001
                gl._shadow_window.switch_to()  # noqa: SLF001

    def _safe_to_operate_on_object_space(self) -> bool:
        """Check if it's safe to interact with this context's object space.

        This is considered to be the case if the currently active context's
        object space is the same as this context's object space and this
        method is called from the main thread.
        """
        return (self.object_space is gl.current_context.object_space and
                threading.current_thread() is threading.main_thread())

    def _safe_to_operate_on(self) -> bool:
        """Check whether it is safe to interact with this context.

        This is considered to be the case if it's the current context and this
        method is called from the main thread.
        """
        return gl.current_context is self and threading.current_thread() is threading.main_thread()

    def create_program(self, *sources: tuple[str, str]) -> ShaderProgram:
        """Create a ShaderProgram from OpenGL GLSL source.

        This is a convenience method that takes one or more tuples of
        (source_string, shader_type), and returns a
        :py:class:`~pyglet.graphics.shader.ShaderProgram` instance.

        ``source_string`` is OpenGL GLSL source code as a str, and ``shader_type``
        is the OpenGL shader type, such as "vertex" or "fragment". See
        :py:class:`~pyglet.graphics.shader.Shader` for more information.

        .. note:: This method is cached. Given the same shader sources, the
                  same ShaderProgram instance will be returned. For more
                  control over the ShaderProgram lifecycle, it is recommended
                  to manually create Shaders and link ShaderPrograms.

        .. versionadded:: 2.0.10
        """
        if program := self._cached_programs.get(str(sources)):
            return program

        shaders = (pyglet.graphics.shader.Shader(src, srctype) for (src, srctype) in sources)
        program = pyglet.graphics.shader.ShaderProgram(*shaders)
        self._cached_programs[str(sources)] = program

        return program

    def delete_texture(self, texture_id: int) -> None:
        """Safely delete a Texture belonging to this context's object space.

        This method will delete the texture immediately via
        ``glDeleteTextures`` if the current context's object space is the same
        as this context's object space, and it is called from the main thread.

        Otherwise, the texture will only be marked for deletion, postponing
        it until any context with the same object space becomes active again.

        This makes it safe to call from anywhere, including other threads.
        """
        if self._safe_to_operate_on_object_space():
            gl.glDeleteTextures(1, gl.GLuint(texture_id))
        else:
            self.object_space.doomed_textures.append(texture_id)

    def delete_buffer(self, buffer_id: int) -> None:
        """Safely delete a Buffer belonging to this context's object space.

        This method behaves similarly to ``delete_texture``, though for
        ``glDeleteBuffers`` instead of ``glDeleteTextures``.
        """
        if self._safe_to_operate_on_object_space():
            gl.glDeleteBuffers(1, gl.GLuint(buffer_id))
        else:
            self.object_space.doomed_buffers.append(buffer_id)

    def delete_shader_program(self, program_id: int) -> None:
        """Safely delete a ShaderProgram belonging to this context's object space.

        This method behaves similarly to ``delete_texture``, though for
        ``glDeleteProgram`` instead of ``glDeleteTextures``.
        """
        if self._safe_to_operate_on_object_space():
            gl.glDeleteProgram(gl.GLuint(program_id))
        else:
            self.object_space.doomed_shader_programs.append(program_id)

    def delete_shader(self, shader_id: int) -> None:
        """Safely delete a Shader belonging to this context's object space.

        This method behaves similarly to ``delete_texture``, though for
        ``glDeleteShader`` instead of ``glDeleteTextures``.
        """
        if self._safe_to_operate_on_object_space():
            gl.glDeleteShader(gl.GLuint(shader_id))
        else:
            self.object_space.doomed_shaders.append(shader_id)

    def delete_renderbuffer(self, rbo_id: int) -> None:
        """Safely delete a Renderbuffer belonging to this context's object space.

        This method behaves similarly to ``delete_texture``, though for
        ``glDeleteRenderbuffers`` instead of ``glDeleteTextures``.
        """
        if self._safe_to_operate_on_object_space():
            gl.glDeleteRenderbuffers(1, gl.GLuint(rbo_id))
        else:
            self.object_space.doomed_renderbuffers.append(rbo_id)

    def delete_vao(self, vao_id: int) -> None:
        """Safely delete a Vertex Array Object belonging to this context.

        If this context is not the current context or this method is not
        called from the main thread, its deletion will be postponed until
        this context is next made active again.

        Otherwise, this method will immediately delete the VAO via
        ``glDeleteVertexArrays``.
        """
        if self._safe_to_operate_on():
            gl.glDeleteVertexArrays(1, gl.GLuint(vao_id))
        else:
            self.doomed_vaos.append(vao_id)

    def delete_framebuffer(self, fbo_id: int) -> None:
        """Safely delete a Framebuffer Object belonging to this context.

        This method behaves similarly to ``delete_vao``, though for
        ``glDeleteFramebuffers`` instead of ``glDeleteVertexArrays``.
        """
        if self._safe_to_operate_on():
            gl.glDeleteFramebuffers(1, gl.GLuint(fbo_id))
        else:
            self.doomed_framebuffers.append(fbo_id)

    def get_info(self) -> GLInfo:
        """Get the :py:class:`~GLInfo` instance for this context."""
        return self._info



"""Formatted and unformatted document interfaces used by text layout.

Abstract representation
=======================

Styled text in pyglet is represented by one of the :py:class:`~pyglet.text.document.AbstractDocument` classes,
which manage the state representation of text and style independently of how
it is loaded or rendered.

A document consists of the document text (a Unicode string) and a set of
named style ranges.  For example, consider the following (artificial)
example::

    0    5   10   15   20
    The cat sat on the mat.
    +++++++        +++++++    "weight"
                ++++++      "italic"

If this example were to be rendered, "The cat" and "the mat" would have a weight
set ("bold", "thin", etc.), and "on the" would be in italics.  Note that the second
"the" is both weighted and italic.

The document styles recorded for this example would be a specific ``"weight"`` over
ranges (0-7) & (15-22) and ``"italic"`` over range (12-18).  Overlapping styles are
permitted; unlike HTML and other structured markup, the ranges need not be nested.

The document has no knowledge of the semantics of ``"weight"`` names or
``"italic"``, it stores only the style names.  The pyglet layout classes give
meaning to these style names in the way they are rendered; but you are also free
to invent your own style names (which will be ignored by the layout classes).
This can be useful to tag areas of interest in a document, or maintain
references back to the source material.

As well as text, the document can contain arbitrary elements represented by
:py:class:`~pyglet.text.document.InlineElement`.  An inline element behaves
like a single character in the document, but can be rendered by the application.

Paragraph breaks
================

Paragraph breaks are marked with a "newline" character (U+0010).  The Unicode
paragraph break (U+2029) can also be used.

Line breaks (U+2028) can be used to force a line break within a paragraph.

See Unicode recommendation UTR #13 for more information:
https://www.unicode.org/standard/reports/tr13/tr13-5.html.

Document classes
================

Any class implementing :py:class:`~pyglet.text.document.AbstractDocument` provides an interface to a
document model as described above.  In theory a structured document such as
HTML or XML could export this model, though the classes provided by pyglet
implement only unstructured documents.

The :py:class:`~pyglet.text.document.UnformattedDocument` class assumes any styles set are set over the entire
document.  So, regardless of the range specified when setting a ``"weight"``
style attribute, for example, the entire document will receive that style.

The :py:class:`~pyglet.text.document.FormattedDocument` class implements the document model directly, using
the `RunList` class to represent style runs efficiently.

Style attributes
================

The following character style attribute names are recognised by pyglet:

``font_name``
    Font family name, as given to :py:func:`pyglet.font.load`.
``font_size``
    Font size, in points.
``weight``
    String.
``italic``
    Boolean.
``underline``
    4-tuple of ints in range (0, 255) giving RGBA underline color, or None
    (default) for no underline.
``kerning``
    Additional space to insert between glyphs, in points.  Defaults to 0.
``baseline``
    Offset of glyph baseline from line baseline, in points.  Positive values
    give a superscript, negative values give a subscript.  Defaults to 0.
``color``
    4-tuple of ints in range (0, 255) giving RGBA text color
``background_color``
    4-tuple of ints in range (0, 255) giving RGBA text background color; or
    ``None`` for no background fill.

The following paragraph style attribute names are recognised by pyglet.  Note
that paragraph styles are handled no differently from character styles by the
document: it is the application's responsibility to set the style over an
entire paragraph, otherwise results are undefined.

``align``
    ``left`` (default), ``center`` or ``right``.
``indent``
    Additional horizontal space to insert before the first
``leading``
    Additional space to insert between consecutive lines within a paragraph,
    in points.  Defaults to 0.
``line_spacing``
    Distance between consecutive baselines in a paragraph, in points.
    Defaults to ``None``, which automatically calculates the tightest line
    spacing for each line based on the font ascent and descent.
``margin_left``
    Left paragraph margin, in pixels.
``margin_right``
    Right paragraph margin, in pixels.
``margin_top``
    Margin above paragraph, in pixels.
``margin_bottom``
    Margin below paragraph, in pixels.  Adjacent margins do not collapse.
``tab_stops``
    List of horizontal tab stops, in pixels, measured from the left edge of
    the text layout.  Defaults to the empty list.  When the tab stops
    are exhausted, they implicitly continue at 50 pixel intervals.
``wrap``
    Boolean.  If True (the default), text wraps within the width of the layout.

Other attributes can be used to store additional style information within the
document; it will be ignored by the built-in text classes.

All style attributes (including those not present in a document) default to
``None`` (including the so-called "boolean" styles listed above).  The meaning
of a ``None`` style is style- and application-dependent.
"""
from __future__ import annotations

import re
import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generator

from pyglet import event
from pyglet.text import runlist

if TYPE_CHECKING:
    from pyglet.font.base import Font
    from pyglet.text.layout import TextLayout

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run

#: The style attribute takes on multiple values in the document.
STYLE_INDETERMINATE = "indeterminate"


class InlineElement(ABC):
    """Arbitrary inline element positioned within a formatted document.

    Elements behave like a single glyph in the document.  They are
    measured by their horizontal advance, ascent above the baseline, and
    descent below the baseline.

    The pyglet layout classes reserve space in the layout for elements and
    call the element's methods to ensure they are rendered at the
    appropriate position.

    If the size of an element (any of the ``advance``, ``ascent``, or ``descent``
    variables) is modified it is the application's responsibility to
    trigger a reflow of the appropriate area in the affected layouts.  This
    can be done by forcing a style change over the element's position.
    """
    advance: int
    descent: int
    ascent: int
    _position: int | None

    def __init__(self, ascent: int, descent: int, advance: int) -> None:
        """Initialize the element.

        Args:
            ascent:
                Ascent of the element above the baseline, in pixels.
            descent:
                Descent of the element below the baseline, in pixels. Typically negative.
            advance:
                Width of the element, in pixels.
        """
        self.ascent = ascent
        self.descent = descent
        self.advance = advance
        self._position = None

    @property
    def position(self) -> int | None:
        """Character position within the document.

        Determined by the layout it is in. Will return ``None`` if it has not been placed.
        """
        return self._position

    @abstractmethod
    def place(self, layout: TextLayout, x: float, y: float, z: float, line_x: float, line_y: float, rotation: float,
              visible: bool, anchor_x: float, anchor_y: float) -> None:
        ...

    @abstractmethod
    def update_translation(self, x: float, y: float, z: float) -> None:
        ...

    @abstractmethod
    def update_color(self, color: list[int]) -> None:
        ...

    @abstractmethod
    def update_view_translation(self, translate_x: float, translate_y: float) -> None:
        ...

    @abstractmethod
    def update_rotation(self, rotation: float) -> None:
        ...

    @abstractmethod
    def update_visibility(self, visible: bool) -> None:
        ...

    @abstractmethod
    def update_anchor(self, anchor_x: float, anchor_y: float) -> None:
        ...

    @abstractmethod
    def remove(self, layout: TextLayout) -> None:
        ...


class AbstractDocument(event.EventDispatcher):
    """Abstract document interface used by all :py:mod:`pyglet.text` classes.

    This class can be overridden to interface pyglet with a third-party
    document format.  It may be easier to implement the document format in
    terms of one of the supplied concrete classes :py:class:`~pyglet.text.document.FormattedDocument` or
    :py:class:`~pyglet.text.document.UnformattedDocument`.
    """
    _previous_paragraph_re: re.Pattern[str] = re.compile("\n[^\n\u2029]*$")
    _next_paragraph_re: re.Pattern[str] = re.compile("[\n\u2029]")

    def __init__(self, text: str = "") -> None:
        """Initialize a document with text.

        Args:
            text: Initial text string.
        """
        super().__init__()

        self._text = ""
        self._elements: list[InlineElement] = []
        if text:
            self.insert_text(0, text)

    @property
    def text(self) -> str:
        """Document text.

        For efficient incremental updates, use the :py:func:`~pyglet.text.document.AbstractDocument.insert_text` and
        :py:func:`~pyglet.text.document.AbstractDocument.delete_text` methods instead of replacing this property.

        """
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        if text == self._text:
            return
        self.delete_text(0, len(self._text))
        self.insert_text(0, text)

    def get_paragraph_start(self, pos: int) -> int:
        """Get the starting position of a paragraph from the character position."""
        # Tricky special case where the $ in pattern matches before the
        # \n at the end of the string instead of the end of the string.
        if self._text[:pos + 1].endswith("\n") or self._text[:pos + 1].endswith("\u2029"):
            return pos

        m: re.Match[str] | None = self._previous_paragraph_re.search(self._text, 0, pos + 1)
        if not m:
            return 0
        return m.start() + 1

    def get_paragraph_end(self, pos: int) -> int:
        """Get the end position of a paragraph from the character position."""
        m: re.Match[str] | None = self._next_paragraph_re.search(self._text, pos)
        if not m:
            return len(self._text)
        return m.start() + 1

    @abstractmethod
    def get_style_runs(self, attribute: str) -> runlist.AbstractRunIterator:
        """Get a style iterator over the given style attribute.

        Args:
            attribute:
                Name of style attribute to query.
        """

    @abstractmethod
    def get_style(self, attribute: str, position: int = 0) -> Any:
        """Get an attribute style at the given position.

        Args:
            attribute:
                Name of style attribute to query.
            position:
                Character position of document to query.

        Returns:
            The style set for the attribute at the given position.
        """

    def get_style_range(self, attribute: str, start: int, end: int) -> Any:
        """Get an attribute style over the given range.

        If the style varies over the range, `STYLE_INDETERMINATE` is returned.

        Args:
            attribute:
                Name of style attribute to query.
            start:
                Starting character position.
            end:
                Ending character position (exclusive).

        Returns:
            The style set for the attribute over the given range, or `STYLE_INDETERMINATE` if more than one value is
            set.
        """
        iterable: runlist.AbstractRunIterator = self.get_style_runs(attribute)
        _, value_end, value = next(iterable.ranges(start, end))
        if value_end < end:
            return STYLE_INDETERMINATE

        return value

    @abstractmethod
    def get_font_runs(self, dpi: int | None = None) -> runlist.AbstractRunIterator:
        """Get a style iterator over the `pyglet.font.Font` instances used in the document.

        The font instances are created on-demand by inspection of the
        ``font_name``, ``font_size``, ``weight`` and ``italic`` style
        attributes.

        Args:
            dpi:
                Optional resolution to construct fonts at.
                :see: :py:func:`~pyglet.font.load`.
        """

    @abstractmethod
    def get_font(self, position: int, dpi: int | None = None) -> Font:
        """Get the font instance used at the given position.

        :see: `get_font_runs`

        Args:
            position:
                Character position of document to query.
            dpi:
                Optional resolution to construct fonts at.
                :see: :py:func:`~pyglet.font.load`.
        """

    def insert_text(self, start: int, text: str, attributes: dict[str, Any] | None = None) -> None:  # noqa: D417
        """Insert text into the document.

        Dispatches an :py:meth:`~pyglet.text.document.AbstractDocument.on_insert_text` event.

        Args:
            start:
                Character insertion point within document.
            text:
                Text to insert.
            attributes:
                Optional dictionary giving named style attributes of the inserted text.
        """  # noqa: D411, D405, D214, D410
        self._insert_text(start, text, attributes)
        self.dispatch_event("on_insert_text", start, text)

    def _insert_text(self, start: int, text: str, attributes: dict[str, Any] | None) -> None:  # noqa: ARG002
        self._text = "".join((self._text[:start], text, self._text[start:]))
        len_text = len(text)
        for element in self._elements:
            assert element._position is not None  # noqa: SLF001
            if element._position >= start:  # noqa: SLF001
                element._position += len_text  # noqa: SLF001

    def delete_text(self, start: int, end: int) -> None:
        """Delete text from the document.

        Dispatches an :py:meth:`on_delete_text` event.

        Args:
            start:
                Starting character position to delete from.
            end:
                Ending character position to delete to (exclusive).

        """
        self._delete_text(start, end)
        self.dispatch_event("on_delete_text", start, end)

    def _delete_text(self, start: int, end: int) -> None:
        for element in list(self._elements):
            assert element.position is not None
            if start <= element._position < end: # noqa: SLF001
                self._elements.remove(element)
            elif element._position >= end:  # fixes #538  # noqa: SLF001
                element._position -= (end - start)  # noqa: SLF001

        self._text = self._text[:start] + self._text[end:]

    def insert_element(self, position: int, element: InlineElement,
                       attributes: dict[str, Any] | None = None) -> None:
        """Insert a element into the document.

        See the :py:class:`~pyglet.text.document.InlineElement` class documentation for details of usage.

        Args:
            position:
                Character insertion point within document.
            element:
                Element to insert.
            attributes: dict
                Optional dictionary giving named style attributes of the inserted text.

        """
        assert element.position is None, "Element is already in a document."
        self.insert_text(position, "\0", attributes)
        element._position = position  # noqa: SLF001
        self._elements.append(element)

        # All _elements should have a valid position assigned.
        self._elements.sort(key=lambda d: d.position)  # type: ignore[arg-type, return-value]

    def get_element(self, position: int) -> InlineElement:
        """Get the element at a specified position.

        Args:
            position:
                Position in the document of the element.

        """
        for element in self._elements:
            if element._position == position:  # noqa: SLF001
                return element
        msg = f"No element at position {position}"
        raise RuntimeError(msg)

    def set_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:  # noqa: D417
        """Set text style of a range between start and end of the document.

        Dispatches an :py:meth:`~pyglet.text.document.AbstractDocument.on_style_text` event.

        Args:
            start:
                Starting character position.
            end:
                Ending character position (exclusive).
            attributes:
                Dictionary giving named style attributes of the text.

        """  # noqa: D214, D405, D411, D410
        self._set_style(start, end, attributes)
        self.dispatch_event("on_style_text", start, end, attributes)

    @abstractmethod
    def _set_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:
        ...

    def set_paragraph_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:  # noqa: D417
        """Set the style for a range of paragraphs.

        This is a convenience method for `set_style` that aligns the character range to the enclosing paragraph(s).

        Dispatches an :py:meth:`~pyglet.text.document.AbstractDocument.on_style_text` event.

        Args:
            start:
                Starting character position.
            end:
                Ending character position (exclusive).
            attributes:
                Dictionary giving named style attributes of the paragraphs.

        """  # noqa: D214, D405, D411, D410
        start = self.get_paragraph_start(start)
        end = self.get_paragraph_end(end)
        self._set_style(start, end, attributes)
        self.dispatch_event("on_style_text", start, end, attributes)

    if _is_pyglet_doc_run:
        def on_insert_text(self, start: int, text: str) -> None:
            """Text was inserted into the document.

            Args:
                start:
                    Character insertion point within document.
                text:
                    The text that was inserted.

            :event:
            """

        def on_delete_text(self, start: int, end: int) -> None:
            """Text was deleted from the document.

            Args:
                start:
                    Starting character position of deleted text.
                end:
                    Ending character position of deleted text (exclusive).

            :event:
            """

        def on_style_text(self, start: int, end: int, attributes: dict[str, Any] | None) -> None:  # noqa: D417
            """Text character style was modified.

            Args:
                start:
                    Starting character position of modified text.
                end:
                    Ending character position of modified text (exclusive).
                attributes:
                    Dictionary giving updated named style attributes of the text.

            :event:
            """  # noqa: D214, D405, D411, D410


AbstractDocument.register_event_type("on_insert_text")
AbstractDocument.register_event_type("on_delete_text")
AbstractDocument.register_event_type("on_style_text")


class UnformattedDocument(AbstractDocument):
    """A document having uniform style over all text.

    Changes to the style of text within the document affects the entire
    document.  For convenience, the ``position`` parameters of the style
    methods may therefore be omitted.
    """

    def __init__(self, text: str = "") -> None:
        """Create unformatted document with a string."""
        super().__init__(text)
        self.styles: dict[str, Any] = {}

    def get_style_runs(self, attribute: str) -> runlist.ConstRunIterator:
        value = self.styles.get(attribute)
        return runlist.ConstRunIterator(len(self.text), value)

    def get_style(self, attribute: str, position: int | None = None) -> Any:  # noqa: ARG002
        return self.styles.get(attribute)

    def set_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:  # noqa: ARG002
        return super().set_style(0, len(self.text), attributes)

    def _set_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:  # noqa: ARG002
        self.styles.update(attributes)

    def set_paragraph_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:  # noqa: ARG002
        return super().set_paragraph_style(0, len(self.text), attributes)

    def get_font_runs(self, dpi: int | None = None) -> runlist.ConstRunIterator:
        ft: Font = self.get_font(dpi=dpi)
        return runlist.ConstRunIterator(len(self.text), ft)

    def get_font(self, position: int | None = None, dpi: int | None = None) -> Font:  # noqa: ARG002
        from pyglet import font
        font_name = self.styles.get("font_name")
        font_size = self.styles.get("font_size")
        weight = self.styles.get("weight", "normal")
        italic = self.styles.get("italic", False)
        stretch = self.styles.get("stretch", False)
        return font.load(font_name, font_size, weight=weight, italic=italic, stretch=stretch, dpi=dpi)

    def get_element_runs(self) -> runlist.ConstRunIterator:
        return runlist.ConstRunIterator(len(self._text), None)


class FormattedDocument(AbstractDocument):
    """Simple implementation of a document that maintains text formatting.

    Changes to text style are applied according to the description in
    :py:class:`~pyglet.text.document.AbstractDocument`.  All styles default to ``None``.
    """

    def __init__(self, text: str = "") -> None:  # noqa: D107
        self._style_runs: dict[str, runlist.RunList] = {}
        super().__init__(text)

    def get_style_runs(self, attribute: str) -> runlist.RunIterator | _NoStyleRangeIterator:
        try:
            return self._style_runs[attribute].get_run_iterator()
        except KeyError:
            return _no_style_range_iterator

    def get_style(self, attribute: str, position: int = 0) -> Any | None:
        try:
            return self._style_runs[attribute][position]
        except KeyError:
            return None

    def _set_style(self, start: int, end: int, attributes: dict[str, Any]) -> None:
        for attribute, value in attributes.items():
            try:
                runs = self._style_runs[attribute]
            except KeyError:
                runs = self._style_runs[attribute] = runlist.RunList(0, None)
                runs.insert(0, len(self._text))
            runs.set_run(start, end, value)

    def get_font_runs(self, dpi: int | None = None) -> _FontStyleRunsRangeIterator:
        return _FontStyleRunsRangeIterator(
            self.get_style_runs("font_name"),
            self.get_style_runs("font_size"),
            self.get_style_runs("weight"),
            self.get_style_runs("italic"),
            self.get_style_runs("stretch"),
            dpi)

    def get_font(self, position: int, dpi: int | None = None) -> Font:
        runs_iter = self.get_font_runs(dpi)
        return runs_iter[position]

    def get_element_runs(self) -> _ElementIterator:
        return _ElementIterator(self._elements, len(self._text))

    def _insert_text(self, start: int, text: str, attributes: dict[str, Any]) -> None:
        super()._insert_text(start, text, attributes)

        len_text = len(text)
        for runs in self._style_runs.values():
            runs.insert(start, len_text)

        if attributes is not None:
            for attribute, value in attributes.items():
                try:
                    runs = self._style_runs[attribute]
                except KeyError:
                    runs = self._style_runs[attribute] = runlist.RunList(0, None)
                    runs.insert(0, len(self.text))
                runs.set_run(start, start + len_text, value)

    def _delete_text(self, start: int, end: int) -> None:
        super()._delete_text(start, end)
        for runs in self._style_runs.values():
            runs.delete(start, end)


def _iter_elements(elements: list[InlineElement], length: int) -> Generator[
    tuple[int, int, InlineElement | None], None, None]:
    last = 0
    for element in elements:
        p = element.position
        assert p is not None
        yield last, p, None
        yield p, p + 1, element
        last = p + 1
    yield last, length, None


class _ElementIterator(runlist.RunIterator):
    def __init__(self, elements: list[InlineElement], length: int) -> None:
        self._run_list_iter = _iter_elements(elements, length)
        self.start, self.end, self.value = next(self)


class _FontStyleRunsRangeIterator(runlist.RunIterator):
    # XXX subclass runlist
    def __init__(self, font_names: runlist.RunIterator, font_sizes: runlist.RunIterator, weights: runlist.RunIterator,
                 italics: runlist.RunIterator, stretch: runlist.RunIterator, dpi: int | None) -> None:
        self.zip_iter = runlist.ZipRunIterator((font_names, font_sizes, weights, italics, stretch))
        self.dpi = dpi

    def ranges(self, start: int, end: int) -> Generator[tuple[int, int, Font], None, None]:
        from pyglet import font
        for start_, end_, styles in self.zip_iter.ranges(start, end):
            font_name, font_size, weight, italic, stretch = styles
            ft = font.load(font_name, font_size, weight=weight, italic=bool(italic), stretch=stretch, dpi=self.dpi)
            yield start_, end_, ft

    def __getitem__(self, index: int) -> Font:
        from pyglet import font
        font_name, font_size, weight, italic, stretch = self.zip_iter[index]
        return font.load(font_name, font_size, weight=weight, italic=bool(italic), stretch=stretch, dpi=self.dpi)


class _NoStyleRangeIterator(runlist.RunIterator):
    def __init__(self) -> None:
        pass

    def ranges(self, start: int, end: int) -> Generator[tuple[int, int, None], None, None]:
        yield start, end, None

    def __getitem__(self, index: int) -> None:
        return None


_no_style_range_iterator = _NoStyleRangeIterator()



import pyglet

from pyglet.gl import glEnable, GL_DEPTH_TEST, GL_CULL_FACE
from pyglet.math import Mat4, Vec3


window = pyglet.window.Window(resizable=True)
batch = pyglet.graphics.Batch()
time = 0


@window.event
def on_resize(width, height):
    window.viewport = (0, 0, width, height)

    window.projection = Mat4.perspective_projection(window.aspect_ratio, z_near=0.1, z_far=255)
    return pyglet.event.EVENT_HANDLED


@window.event
def on_draw():
    window.clear()
    batch.draw()


def animate(dt):
    global time
    time += dt

    rot_x = Mat4.from_rotation(time, Vec3(1, 0, 0))
    rot_y = Mat4.from_rotation(time/2, Vec3(0, 1, 0))
    rot_z = Mat4.from_rotation(time/3, Vec3(0, 0, 1))
    trans = Mat4.from_translation(Vec3(1.25, 0, 2))
    model_logo.matrix = trans @ rot_x @ rot_y @ rot_z

    rot_x = Mat4.from_rotation(time, Vec3(1, 0, 0))
    rot_y = Mat4.from_rotation(time/3, Vec3(0, 1, 0))
    rot_z = Mat4.from_rotation(time/2, Vec3(0, 0, 1))
    trans = Mat4.from_translation(Vec3(-1.75, 0, 0))
    model_box.matrix = trans @ rot_x @ rot_y @ rot_z


if __name__ == "__main__":
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_CULL_FACE)

    logo_scene = pyglet.resource.scene('logo3d.obj')
    box_scene = pyglet.resource.scene('box.obj')
    model_logo = logo_scene.create_models(batch=batch)[0]   # only one model in this scene
    model_box = box_scene.create_models(batch=batch)[0]     # only one model in this scene

    # Set the application wide view matrix (camera):
    window.view = Mat4.look_at(position=Vec3(0, 0, 5), target=Vec3(0, 0, 0), up=Vec3(0, 1, 0))

    pyglet.clock.schedule_interval(animate, 1 / 60)
    pyglet.app.run()



"""Wrapper for include/libavutil/avutil.h
"""
from ctypes import c_char_p, c_void_p, POINTER, Structure, c_double, c_char, Union
from ctypes import c_int, c_int64, c_uint64
from ctypes import c_uint8, c_int8, c_uint, c_size_t

import pyglet.lib
from pyglet.util import debug_print
from . import compat

_debug = debug_print('debug_media')

avutil = pyglet.lib.load_library(
    'avutil',
    win32=('avutil-59', 'avutil-58', 'avutil-57', 'avutil-56'),
    darwin=('avutil.59', 'avutil.58', 'avutil.57', 'avutil.56')
)

avutil.avutil_version.restype = c_int

avutil_version = avutil.avutil_version() >> 16

compat.set_version('avutil', avutil_version)

AVMEDIA_TYPE_UNKNOWN = -1
AVMEDIA_TYPE_VIDEO = 0
AVMEDIA_TYPE_AUDIO = 1
AVMEDIA_TYPE_DATA = 2
AVMEDIA_TYPE_SUBTITLE = 3
AVMEDIA_TYPE_ATTACHMENT = 4
AVMEDIA_TYPE_NB = 5

AV_SAMPLE_FMT_U8 = 0
AV_SAMPLE_FMT_S16 = 1
AV_SAMPLE_FMT_S32 = 2
AV_SAMPLE_FMT_FLT = 3
AV_SAMPLE_FORMAT_DOUBLE = 4
AV_SAMPLE_FMT_U8P = 5
AV_SAMPLE_FMT_S16P = 6
AV_SAMPLE_FMT_S32P = 7
AV_SAMPLE_FMT_FLTP = 8
AV_SAMPLE_FMT_DBLP = 9
AV_SAMPLE_FMT_S64 = 10
AV_SAMPLE_FMT_S64P = 11

AV_NUM_DATA_POINTERS = 8

AV_PIX_FMT_RGB24 = 2
AV_PIX_FMT_ARGB = 25
AV_PIX_FMT_RGBA = 26

AVChannel = c_int
AVChannelOrder = c_int

class AVChannelCustom(Structure):
    _fields_ = [
        ('id', AVChannel),
        ('name', c_char * 16),
        ('opaque', c_void_p)
    ]

class _AVChannelLayoutUnion(Union):
    _fields_ = [
        ('mask', c_uint64),
        ('map', POINTER(AVChannelCustom)),
    ]

class AVChannelLayout(Structure):
    _fields_ = [
        ('order', c_int),
        ('nb_channels', c_int),
        ('u', _AVChannelLayoutUnion),
        ('opaque', c_void_p)
    ]

    def __repr__(self):
        return f"{self.__class__.__name__}(order={self.order}, nb_channels={self.nb_channels})"

class AVBuffer(Structure):
    _fields_ = [
        ('data', POINTER(c_uint8)),
        ('size', c_int),
        # .. more
    ]


class AVBufferRef(Structure):
    _fields_ = [
        ('buffer', POINTER(AVBuffer)),
        ('data', POINTER(c_uint8)),
        ('size', c_int)
    ]


class AVDictionaryEntry(Structure):
    _fields_ = [
        ('key', c_char_p),
        ('value', c_char_p)
    ]


class AVDictionary(Structure):
    _fields_ = [
        ('count', c_int),
        ('elems', POINTER(AVDictionaryEntry))
    ]


class AVClass(Structure):
    pass


class AVRational(Structure):
    _fields_ = [
        ('num', c_int),
        ('den', c_int)
    ]

    def __repr__(self):
        return f"AVRational({self.num}/{self.den})"


class AVFrameSideData(Structure):
    pass


class AVFrame(Structure):
    pass

AVFrame_Fields = [
    ('data', POINTER(c_uint8) * AV_NUM_DATA_POINTERS),
    ('linesize', c_int * AV_NUM_DATA_POINTERS),
    ('extended_data', POINTER(POINTER(c_uint8))),
    ('width', c_int),
    ('height', c_int),
    ('nb_samples', c_int),
    ('format', c_int),
    ('key_frame', c_int),  # Deprecated in 59
    ('pict_type', c_int),
    ('sample_aspect_ratio', AVRational),
    ('pts', c_int64),
    ('pkt_pts', c_int64),  # Deprecated. Removed in 57.
    ('pkt_dts', c_int64),
    ('time_base', AVRational),  # Added (5.x)+
    ('coded_picture_number', c_int),  # removed in 59
    ('display_picture_number', c_int),  # removed in 59
    ('quality', c_int),
    ('opaque', c_void_p),
    ('error', c_uint64 * AV_NUM_DATA_POINTERS),  # Deprecated. Removed in 57.
    ('repeat_pict', c_int),
    ('interlaced_frame', c_int),  # deprecated in 59. Targeted for removal. Use AV_FRAME_FLAG_INTERLACED
    ('top_field_first', c_int),  # deprecated in 59. Targeted for removal. Use AV_FRAME_FLAG_TOP_FIELD_FIRST
    ('palette_has_changed', c_int),  # deprecated in 59. Targeted for removal.
    ('reordered_opaque', c_int64),  # removed in 59.
    ('sample_rate', c_int),
    ('channel_layout', c_uint64),  # removed in 59.
    ('buf', POINTER(AVBufferRef) * AV_NUM_DATA_POINTERS),
    ('extended_buf', POINTER(POINTER(AVBufferRef))),
    ('nb_extended_buf', c_int),
    ('side_data', POINTER(POINTER(AVFrameSideData))),
    ('nb_side_data', c_int),
    ('flags', c_int),
    ('color_range', c_int),
    ('color_primaries', c_int),
    ('color_trc', c_int),
    ('colorspace', c_int),
    ('chroma_location', c_int),
    ('best_effort_timestamp', c_int64),
    ('pkt_pos', c_int64),  # deprecated in 59. Use AV_CODEC_FLAG_COPY_OPAQUE
    ('pkt_duration', c_int64),  # removed in 59?
    ('metadata', POINTER(AVDictionary)),
    ('decode_error_flags', c_int),
    ('channels', c_int),
    ('pkt_size', c_int),  # deprecated in 59.  use AV_CODEC_FLAG_COPY_OPAQUE to pass through arbitrary user data from packets to frames
    ('qscale_table', POINTER(c_int8)),  # Deprecated. Removed in 57.
    ('qstride', c_int),  # Deprecated. Removed in 57.
    ('qscale_type', c_int),  # Deprecated. Removed in 57.
    ('qp_table_buf', POINTER(AVBufferRef)),  # Deprecated. Removed in 57.
    ('hw_frames_ctx', POINTER(AVBufferRef)),
    ('opaque_ref', POINTER(AVBufferRef)),
    ('crop_top', c_size_t),  # video frames only
    ('crop_bottom', c_size_t),  # video frames only
    ('crop_left', c_size_t),  # video frames only
    ('crop_right', c_size_t),  # video frames only
    ('private_ref', POINTER(AVBufferRef)),
    ('ch_layout', AVChannelLayout),
    ('duration', c_int64)
]

compat.add_version_changes('avutil', 56, AVFrame, AVFrame_Fields,
                           removals=('time_base', 'ch_layout', 'duration'))

for compat_ver in (57, 58):
    compat.add_version_changes('avutil', compat_ver, AVFrame, AVFrame_Fields,
                               removals=('pkt_pts', 'error', 'qscale_table', 'qstride', 'qscale_type', 'qp_table_buf',
                                         'ch_layout', 'duration'))

compat.add_version_changes('avutil', 59, AVFrame, AVFrame_Fields,
                           removals=('pkt_pts', 'error', 'qscale_table', 'qstride', 'qscale_type', 'qp_table_buf',
                                     'channels', 'channel_layout',
                                     'coded_picture_number', 'display_picture_number', 'reordered_opaque', 'pkt_duration'
                                     ))


AV_NOPTS_VALUE = -0x8000000000000000
AV_TIME_BASE = 1000000
AV_TIME_BASE_Q = AVRational(1, AV_TIME_BASE)

avutil.av_version_info.restype = c_char_p
avutil.av_dict_get.restype = POINTER(AVDictionaryEntry)
avutil.av_dict_get.argtypes = [POINTER(AVDictionary),
                               c_char_p, POINTER(AVDictionaryEntry),
                               c_int]
avutil.av_rescale_q.restype = c_int64
avutil.av_rescale_q.argtypes = [c_int64, AVRational, AVRational]
avutil.av_samples_get_buffer_size.restype = c_int
avutil.av_samples_get_buffer_size.argtypes = [POINTER(c_int),
                                              c_int, c_int, c_int]
avutil.av_frame_alloc.restype = POINTER(AVFrame)
avutil.av_frame_free.argtypes = [POINTER(POINTER(AVFrame))]

AVSampleFormat = c_int

if avutil_version <= 57:
    # Removed in 7.x (avutil 58)
    avutil.av_get_default_channel_layout.restype = c_int64
    avutil.av_get_default_channel_layout.argtypes = [c_int]
else:
    # Available in 6.x  (avutil 58, 57)
    avutil.av_channel_layout_default.restype = None
    avutil.av_channel_layout_default.argtypes = [POINTER(AVChannelLayout), c_int]

    avutil.av_channel_layout_uninit.restype = None
    avutil.av_channel_layout_uninit.argtypes = [POINTER(AVChannelLayout)]

    avutil.av_opt_set.restype = c_int
    avutil.av_opt_set.argtypes = [c_void_p, c_char_p, c_char_p, c_int]

    avutil.av_opt_set_int.restype = c_int
    avutil.av_opt_set_int.argtypes = [c_void_p, c_char_p, c_int64, c_int]

    avutil.av_opt_set_double.restype = c_int
    avutil.av_opt_set_double.argtypes = [c_void_p, c_char_p, c_double, c_int]

    avutil.av_opt_set_sample_fmt.restype = c_int
    avutil.av_opt_set_sample_fmt.argtypes = [c_void_p, c_char_p, AVSampleFormat, c_int]

avutil.av_get_bytes_per_sample.restype = c_int
avutil.av_get_bytes_per_sample.argtypes = [c_int]
avutil.av_strerror.restype = c_int
avutil.av_strerror.argtypes = [c_int, c_char_p, c_size_t]

avutil.av_image_fill_arrays.restype = c_int
avutil.av_image_fill_arrays.argtypes = [POINTER(c_uint8) * 4, c_int * 4,
                                        POINTER(c_uint8), c_int, c_int, c_int, c_int]
avutil.av_dict_set.restype = c_int
avutil.av_dict_set.argtypes = [POINTER(POINTER(AVDictionary)),
                               c_char_p, c_char_p, c_int]
avutil.av_dict_free.argtypes = [POINTER(POINTER(AVDictionary))]
avutil.av_log_set_level.restype = c_int
avutil.av_log_set_level.argtypes = [c_uint]
avutil.av_malloc.restype = c_void_p
avutil.av_malloc.argtypes = [c_int]
avutil.av_freep.restype = c_void_p
avutil.av_freep.argtypes = [c_void_p]

__all__ = [
    'avutil',
    'AVMEDIA_TYPE_UNKNOWN',
    'AVMEDIA_TYPE_VIDEO',
    'AVMEDIA_TYPE_AUDIO',
    'AVMEDIA_TYPE_DATA',
    'AVMEDIA_TYPE_SUBTITLE',
    'AVMEDIA_TYPE_ATTACHMENT',
    'AVMEDIA_TYPE_NB',
    'AV_SAMPLE_FMT_U8',
    'AV_SAMPLE_FMT_S16',
    'AV_SAMPLE_FMT_S32',
    'AV_SAMPLE_FMT_FLT',
    'AV_SAMPLE_FORMAT_DOUBLE',
    'AV_SAMPLE_FMT_U8P',
    'AV_SAMPLE_FMT_S16P',
    'AV_SAMPLE_FMT_S32P',
    'AV_SAMPLE_FMT_FLTP',
    'AV_SAMPLE_FMT_DBLP',
    'AV_SAMPLE_FMT_S64',
    'AV_SAMPLE_FMT_S64P',
    'AV_NUM_DATA_POINTERS',
    'AV_PIX_FMT_RGB24',
    'AV_PIX_FMT_ARGB',
    'AV_PIX_FMT_RGBA',
    'AV_NOPTS_VALUE',
    'AV_TIME_BASE',
    'AV_TIME_BASE_Q',
    'AVFrame',
    'AVRational',
    'AVDictionary',
    'AVChannelLayout',
]



from __future__ import annotations

import contextlib
from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    Union,
    c_char_p,
    c_int,
    c_int8,
    c_int16,
    c_int32,
    c_size_t,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint32,
    c_void_p,
)

import pyglet

# Harfbuzz DLL depends on libglib and libintl (for Windows)
hb_lib = None
with contextlib.suppress(ImportError):
    hb_lib = pyglet.lib.load_library("harfbuzz", win32='libharfbuzz-0.dll', darwin='libharfbuzz.0.dylib')


HB_MEMORY_MODE_READONLY = 0  # for read-only font data

HB_DIRECTION_INVALID = 0
HB_DIRECTION_LTR = 4
HB_DIRECTION_RTL = 5
HB_DIRECTION_TTB = 6
HB_DIRECTION_BTT = 7

HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES = 0
HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS = 1
HB_BUFFER_CLUSTER_LEVEL_CHARACTERS = 2
HB_BUFFER_CLUSTER_LEVEL_DEFAULT = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES

HB_GLYPH_FLAG_UNSAFE_TO_BREAK = 0x00000001
HB_GLYPH_FLAG_UNSAFE_TO_CONCAT = 0x00000002
HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL = 0x00000004

HB_GLYPH_FLAG_DEFINED = 0x00000007  # OR of all defined flags

hb_buffer_cluster_level_t = c_uint


class hb_var_int_t(Union):
    _fields_ = [
        ("u32", c_uint32),
        ("i32", c_int32),
        ("u16", c_uint16 * 2),
        ("i16", c_int16 * 2),
        ("u8", c_uint8 * 4),
        ("i8", c_int8 * 4),
    ]


hb_codepoint_t = c_uint32
hb_mask_t = c_uint32


class hb_glyph_info_t(Structure):
    _fields_ = [
        ("codepoint", hb_codepoint_t),  # glyph index
        ("mask", hb_mask_t),
        ("cluster", c_uint32),  # index in the original text
        ("var1", hb_var_int_t),
        ("var2", hb_var_int_t)
    ]

    def __repr__(self) -> str:
        return f"hb_glyph_info_t(codepoint={self.codepoint}, cluster={self.cluster}, flag={self.mask & HB_GLYPH_FLAG_DEFINED})"


hb_position_t = c_int32


# The glyph position structure contains advances and offsets.
class hb_glyph_position_t(Structure):
    _fields_ = [
        ("x_advance", hb_position_t),  # includes kerning
        ("y_advance", hb_position_t),
        ("x_offset", hb_position_t),  # adjustments to the glyph's drawing position
        ("y_offset", hb_position_t),
        ('var', hb_var_int_t),  # Private, do not use, but must be included for proper reading.
    ]


if hb_lib:
    hb_lib.hb_blob_create.argtypes = [
        c_char_p,  # pointer to the font data  const char *data
        c_size_t,  # length of the data
        c_uint,  # memory mode (e.g., HB_MEMORY_MODE_READONLY)  hb_memory_mode_t
        c_void_p,  # user data (None here)  void *user_data
        c_void_p  # destroy callback (None here)  hb_destroy_func_t
    ]
    hb_lib.hb_blob_create.restype = c_void_p

    hb_blob_t = c_void_p
    hb_lib.hb_face_create.argtypes = [
        c_void_p,  # hb_blob_t *blob
        c_uint  # font index in the blob (usually 0)  index
    ]
    hb_lib.hb_face_create.restype = c_void_p

    hb_reference_table_func_t = CFUNCTYPE(c_void_p, c_void_p, c_uint32, c_void_p)
    hb_destroy_func_t = CFUNCTYPE(None, c_void_p)

    hb_lib.hb_face_create_for_tables.argtypes = [
        hb_reference_table_func_t,  # table func
        c_void_p,  # user data ptr
        hb_destroy_func_t  # free memory function
    ]
    hb_lib.hb_face_create_for_tables.restype = c_void_p

    hb_lib.hb_font_create.argtypes = [c_void_p]  # hb_face_t *face
    hb_lib.hb_font_create.restype = c_void_p

    hb_lib.hb_face_get_upem.argtypes = [c_void_p]  # hb_face_t *face
    hb_lib.hb_face_get_upem.restype = c_uint

    hb_lib.hb_font_set_scale.argtypes = [
        c_void_p,  # b_font_t *font
        c_int,  # x_scale
        c_int  # y_scale
    ]
    hb_lib.hb_font_set_scale.restype = None

    hb_lib.hb_buffer_create.argtypes = []
    hb_lib.hb_buffer_create.restype = c_void_p

    hb_lib.hb_buffer_add_utf8.argtypes = [
        c_void_p,  # b_buffer_t *buffer
        c_char_p,  # UTF-8 text  const char *text
        c_int,  # text length (number of bytes)
        c_uint,  # item offset (usually 0)
        c_int  # item length (often the same as text_length)
    ]
    hb_lib.hb_buffer_add_utf8.restype = None

    hb_lib.hb_buffer_set_cluster_level.argtypes = [c_void_p, hb_buffer_cluster_level_t]
    hb_lib.hb_buffer_set_cluster_level.restype = None

    hb_lib.hb_buffer_add_utf16.argtypes = [
        c_void_p,  # hb_buffer_t *buffer
        POINTER(c_uint16),  # UTF-16 text  onst char *text
        c_int,  # text length (number of bytes)
        c_uint,  # item offset (usually 0)
        c_int  # item length (often the same as text_length)
    ]
    hb_lib.hb_buffer_add_utf16.restype = None

    hb_lib.hb_buffer_add_utf32.argtypes = [
        c_void_p,  # hb_buffer_t *buffer
        POINTER(c_uint32),  # UTF-32 text  const char *text
        c_int,  # text length (number of bytes)
        c_uint,  # item offset (usually 0)
        c_int  # item length (often the same as text_length)
    ]
    hb_lib.hb_buffer_add_utf32.restype = None

    hb_lib.hb_buffer_set_direction.argtypes = [
        c_void_p,  # hb_buffer_t *buffer
        c_int  # direction (e.g., HB_DIRECTION_LTR)  hb_direction_t
    ]
    hb_lib.hb_buffer_set_direction.restype = None

    hb_lib.hb_shape.argtypes = [
        c_void_p,  # hb_font_t *font
        c_void_p,  # hb_buffer_t *buffer
        c_void_p,  # hb_feature_t *features (None if not used)
        c_uint  # number of features (0 if features is None)
    ]
    hb_lib.hb_shape.restype = None

    hb_lib.hb_buffer_guess_segment_properties.argtypes = [c_void_p]
    hb_lib.hb_buffer_guess_segment_properties.restype = None

    hb_lib.hb_buffer_get_length.argtypes = [c_void_p]  # hb_buffer_t *buffer
    hb_lib.hb_buffer_get_length.restype = c_uint

    hb_lib.hb_buffer_get_glyph_infos.argtypes = [
        c_void_p,  # hb_buffer_t *buffer,
        POINTER(c_uint)  # pointer for length (can be None)
    ]
    hb_lib.hb_buffer_get_glyph_infos.restype = POINTER(hb_glyph_info_t)

    hb_lib.hb_buffer_get_glyph_positions.argtypes = [
        c_void_p,  # hb_buffer_t *buffer
        POINTER(c_uint)  # pointer for length (can be None)
    ]
    hb_lib.hb_buffer_get_glyph_positions.restype = POINTER(hb_glyph_position_t)

    hb_lib.hb_blob_destroy.argtypes = [c_void_p]  # b_blob_t *blob)
    hb_lib.hb_blob_destroy.restype = None

    hb_lib.hb_face_destroy.argtypes = [c_void_p]  # hb_face_t *face
    hb_lib.hb_face_destroy.restype = None

    hb_lib.hb_font_destroy.argtypes = [c_void_p]  # hb_font_t *font
    hb_lib.hb_font_destroy.restype = None

    hb_lib.hb_buffer_destroy.argtypes = [c_void_p]  # hb_buffer_t *buffer
    hb_lib.hb_buffer_destroy.restype = None

    hb_lib.hb_buffer_normalize_glyphs.argtypes = [c_void_p]
    hb_lib.hb_buffer_normalize_glyphs.restype = None



"""Base class for structured (hierarchical) document formats."""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, NoReturn

import pyglet
import pyglet.text.layout
from pyglet.gl import (
    GL_BLEND,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_SRC_ALPHA,
    GL_TEXTURE0,
    glActiveTexture,
    glBindTexture,
    glBlendFunc,
    glDisable,
    glEnable,
)

if TYPE_CHECKING:
    from pyglet.graphics import Group
    from pyglet.graphics.shader import ShaderProgram
    from pyglet.image import AbstractImage, Texture
    from pyglet.resource import Location
    from pyglet.text.document import InlineElement
    from pyglet.text.layout import TextLayout


class _InlineElementGroup(pyglet.graphics.Group):
    def __init__(self, texture: Texture, program: ShaderProgram, order: int = 1,
                 parent: Group | None = None) -> None:
        super().__init__(order, parent)
        self.texture = texture
        self.program = program

    def set_state(self) -> None:
        self.program.use()

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.texture.target, self.texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def unset_state(self) -> None:
        glDisable(GL_BLEND)
        self.program.stop()

    def __eq__(self, other: _InlineElementGroup) -> bool:
        return (self.__class__ is other.__class__ and
                self._order == other.order and
                self.program == other.program and
                self.parent == other.parent and
                self.texture.target == other.texture.target and
                self.texture.id == other.texture.id)

    def __hash__(self) -> int:
        return hash((self._order, self.program, self.parent,
                     self.texture.target, self.texture.id))


class ImageElement(pyglet.text.document.InlineElement):
    """Adds an image into the layout."""
    height: int
    width: int

    def __init__(self, image: AbstractImage, width: int | None=None, height: int | None=None) -> None:  # noqa: D107
        self.image = image.get_texture()
        self.width = width or image.width
        self.height = height or image.height
        self.vertex_lists = {}

        anchor_y = self.height // image.height * image.anchor_y
        ascent = max(0, self.height - anchor_y)
        descent = min(0, -anchor_y)
        super().__init__(ascent, descent, self.width)

    def place(self, layout: TextLayout, x: float, y: float, z: float, line_x: float, line_y: float, rotation: float,
              visible: bool, anchor_x: float, anchor_y: float) -> None:
        program = pyglet.text.layout.get_default_image_layout_shader()
        group = _InlineElementGroup(self.image.get_texture(), program, 0, layout.group)
        x1 = line_x
        y1 = line_y + self.descent
        x2 = line_x + self.width
        y2 = line_y + self.height + self.descent

        vertex_list = program.vertex_list_indexed(4, pyglet.gl.GL_TRIANGLES, [0, 1, 2, 0, 2, 3],
                                                  layout.batch, group,
                                                  position=("f", (x1, y1, z, x2, y1, z, x2, y2, z, x1, y2, z)),
                                                  translation=("f", (x, y, z) * 4),
                                                  tex_coords=("f", self.image.tex_coords),
                                                  visible=("f", (visible,) * 4),
                                                  rotation=("f", (rotation,) * 4),
                                                  anchor=("f", (anchor_x, anchor_y) * 4),
                                                  )

        self.vertex_lists[layout] = vertex_list

    def update_translation(self, x: float, y: float, z: float) -> None:
        translation = (x, y, z)
        for _vertex_list in self.vertex_lists.values():
            _vertex_list.translation[:] = translation * _vertex_list.count

    def update_color(self, color: list[int]) -> None:
        # No color blending in shader. Optional.
        ...

    def update_view_translation(self, translate_x: float, translate_y: float) -> None:
        view_translation = (-translate_x, -translate_y, 0)
        for _vertex_list in self.vertex_lists.values():
            _vertex_list.view_translation[:] = view_translation * _vertex_list.count

    def update_rotation(self, rotation: float) -> None:
        rot_tuple = (rotation,)
        for _vertex_list in self.vertex_lists.values():
           _vertex_list.rotation[:] = rot_tuple * _vertex_list.count

    def update_visibility(self, visible: bool) -> None:
        visible_tuple = (visible,)
        for _vertex_list in self.vertex_lists.values():
           _vertex_list.visible[:] = visible_tuple * _vertex_list.count

    def update_anchor(self, anchor_x: float, anchor_y: float) -> None:
        anchor = (anchor_x, anchor_y)
        for _vertex_list in self.vertex_lists.values():
            _vertex_list.anchor[:] = anchor * _vertex_list.count

    def remove(self, layout: TextLayout) -> None:
        self.vertex_lists[layout].delete()
        del self.vertex_lists[layout]


def _int_to_roman(number: int) -> str:
    # From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/81611
    if not 0 < number < 4000:
        msg = "Argument must be between 1 and 3999"
        raise ValueError(msg)
    integers = (1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,   4,  1)
    numerals = ("M",  "CM", "D", "CD","C", "XC","L","XL","X","IX","V","IV","I")
    result = ""
    for i in range(len(integers)):
        count = int(number // integers[i])
        result += numerals[i] * count
        number -= integers[i] * count
    return result


class ListBuilder:  # noqa: D101

    def begin(self, decoder: StructuredTextDecoder, style: dict[str, Any]) -> None:
        """Begin a list.

        Args:
            decoder:
                Decoder.
            style:
                Style dictionary that applies over the entire list.
        """
        left_margin = decoder.current_style.get("margin_left") or 0
        tab_stops = decoder.current_style.get("tab_stops")
        if tab_stops:
            tab_stops = list(tab_stops)
        else:
            tab_stops = []
        tab_stops.append(left_margin + 50)
        style["margin_left"] = left_margin + 50
        style["indent"] = -30
        style["tab_stops"] = tab_stops

    def item(self, decoder: StructuredTextDecoder, style: dict[str, Any], value: str | None=None) -> None:  # noqa: ARG002
        """Begin a list item.

        Args:
            decoder:
                `StructuredTextDecoder` Decoder.
            style:
                Style dictionary that applies over the list item.
            value:
                Optional value of the list item.  The meaning is list-type dependent.

        """
        mark = self.get_mark(value)
        if mark:
            decoder.add_text(mark)
        decoder.add_text("\t")

    def get_mark(self, value: str | None=None) -> str:  # noqa: ARG002
        """Get the mark text for the next list item.

        Args:
            value:
                Optional value of the list item.  The meaning is list-type dependent.
        """
        return ""


class UnorderedListBuilder(ListBuilder):  # noqa: D101

    mark: str

    def __init__(self, mark: str) -> None:
        """Create an unordered list with constant mark text.

        Args:
            mark:
                Mark to prepend to each list item.

        """
        self.mark = mark

    def get_mark(self, value: str | None=None) -> str:  # noqa: ARG002
        return self.mark


class OrderedListBuilder(ListBuilder):  # noqa: D101
    format_re = re.compile("(.*?)([1aAiI])(.*)")

    def __init__(self, start: int, fmt: str) -> None:
        """Create an ordered list with sequentially numbered mark text.

        The format is composed of an optional prefix text, a numbering
        scheme character followed by suffix text. Valid numbering schemes
        are:

        ``1``
            Decimal Arabic
        ``a``
            Lowercase alphanumeric
        ``A``
            Uppercase alphanumeric
        ``i``
            Lowercase Roman
        ``I``
            Uppercase Roman

        Prefix text may typically be ``(`` or ``[`` and suffix text is
        typically ``.``, ``)`` or empty, but either can be any string.

        Args:
            start:
                First list item number.
            fmt:
                Format style, for example ``"1."``.

        """
        self.next_value = start

        self.prefix, self.numbering, self.suffix = self.format_re.match(fmt).groups()
        assert self.numbering in "1aAiI"

    def get_mark(self, value: str | None=None) -> str:
        if value is None:
            value = self.next_value
        self.next_value = value + 1
        if self.numbering in "aA":
            try:
                mark = "abcdefghijklmnopqrstuvwxyz"[value - 1]
            except ValueError:
                mark = "?"
            if self.numbering == "A":
                mark = mark.upper()
            return f"{self.prefix}{mark}{self.suffix}"

        if self.numbering in "iI":
            try:
                mark = _int_to_roman(value)
            except ValueError:
                mark = "?"
            if self.numbering == "i":
                mark = mark.lower()
            return f"{self.prefix}{mark}{self.suffix}"

        return f"{self.prefix}{value}{self.suffix}"


class StructuredTextDecoder(pyglet.text.DocumentDecoder):  # noqa: D101
    def decode(self, text: str, location: Location | None=None) -> pyglet.text.document.FormattedDocument:
        self.len_text = 0
        self.current_style = {}
        self.next_style = {}
        self.stack = []
        self.list_stack = []
        self.document = pyglet.text.document.FormattedDocument()
        if location is None:
            location = pyglet.resource.FileLocation("")
        self.decode_structured(text, location)
        return self.document

    def decode_structured(self, text: str, location: Location | None) -> NoReturn:
        raise NotImplementedError

    def push_style(self, key: str, styles: dict[str, Any]) -> None:
        old_styles = {}
        for name in styles:
            old_styles[name] = self.current_style.get(name)
        self.stack.append((key, old_styles))
        self.current_style.update(styles)
        self.next_style.update(styles)

    def pop_style(self, key: str) -> None:
        # Don't do anything if key is not in stack
        for match, _ in self.stack:
            if key == match:
                break
        else:
            return

        # Remove all innermost elements until key is closed.
        while True:
            match, old_styles = self.stack.pop()
            self.next_style.update(old_styles)
            self.current_style.update(old_styles)
            if match == key:
                break

    def add_text(self, text: str) -> None:
        self.document.insert_text(self.len_text, text, self.next_style)
        self.next_style.clear()
        self.len_text += len(text)

    def add_element(self, element: InlineElement) -> None:
        self.document.insert_element(self.len_text, element, self.next_style)
        self.next_style.clear()
        self.len_text += 1



import unittest
from tests import mock
import pyglet.clock


class ClockTestCase(unittest.TestCase):
    """Test clock using dummy time keeper

    not tested:
      positional and named arguments
    """

    def setUp(self):
        self.interval = .001
        self.time = 0
        self.callback_a = mock.Mock()
        self.callback_b = mock.Mock()
        self.callback_c = mock.Mock()
        self.callback_d = mock.Mock()
        self.clock = pyglet.clock.Clock(time_function=lambda: self.time)

    def advance_clock(self, dt=1):
        """simulate the passage of time like a real clock would"""
        frames = 0
        end = self.time + dt
        while self.time < end:
            frames += 1
            self.time += self.interval
            self.clock.tick()
        self.time = round(self.time, 0)
        return frames

    def test_schedule(self):
        self.clock.schedule(self.callback_a)
        frames = self.advance_clock()
        self.assertEqual(self.callback_a.call_count, frames)

    def test_schedule_once(self):
        self.clock.schedule_once(self.callback_a, 1)
        self.advance_clock(2)
        self.assertEqual(self.callback_a.call_count, 1)

    def test_schedule_once_multiple(self):
        self.clock.schedule_once(self.callback_a, 1)
        self.clock.schedule_once(self.callback_b, 2)
        self.advance_clock(2)
        self.assertEqual(self.callback_a.call_count, 1)
        self.assertEqual(self.callback_b.call_count, 1)

    def test_schedule_interval(self):
        self.clock.schedule_interval(self.callback_a, 1)
        self.advance_clock(2)
        self.assertEqual(self.callback_a.call_count, 2)

    def test_schedule_interval_for_duration(self):
        self.clock.schedule_interval_for_duration(self.callback_a, 1, 5)
        self.advance_clock(10)
        self.assertEqual(self.callback_a.call_count, 4)

    def test_schedule_interval_multiple(self):
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.schedule_interval(self.callback_b, 1)
        self.advance_clock(2)
        self.assertEqual(self.callback_a.call_count, 2)
        self.assertEqual(self.callback_b.call_count, 2)

    def test_schedule_interval_soft(self):
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.advance_clock(2)
        self.assertEqual(self.callback_a.call_count, 2)

    @unittest.skip('Requires changes to the clock')
    def test_schedule_interval_soft_multiple(self):
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.schedule_interval_soft(self.callback_b, 1)
        self.clock.schedule_interval_soft(self.callback_b, 1)
        next_ts = set(i.next_ts for i in self.clock._scheduled_items)
        self.assertEqual(len(next_ts), 3)
        self.advance_clock()
        self.assertEqual(self.callback_a.call_count, 1)
        self.assertEqual(self.callback_b.call_count, 2)

    def test_schedule_unschedule(self):
        self.clock.schedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.advance_clock()
        self.assertEqual(self.callback_a.call_count, 0)

    def test_schedule_once_unschedule(self):
        self.clock.schedule_once(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.advance_clock()
        self.assertEqual(self.callback_a.call_count, 0)

    def test_schedule_interval_unschedule(self):
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.advance_clock()
        self.assertEqual(self.callback_a.call_count, 0)

    def test_schedule_interval_soft_unschedule(self):
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.advance_clock()
        self.assertEqual(self.callback_a.call_count, 0)

    def test_unschedule_removes_all(self):
        self.clock.schedule(self.callback_a)
        self.clock.schedule_once(self.callback_a, 1)
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.clock.schedule(self.callback_a)
        self.clock.schedule(self.callback_b)
        self.clock.unschedule(self.callback_a)
        frames = self.advance_clock(10)
        self.assertEqual(self.callback_a.call_count, 0)
        # callback_b is used to verify that the entire event queue was not cleared
        self.assertEqual(self.callback_b.call_count, frames)

    def test_schedule_will_not_call_function(self):
        self.clock.schedule(self.callback_a)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_once(self.callback_a, 0)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_interval(self.callback_a, 1)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.assertEqual(self.callback_a.call_count, 0)

    def test_unschedule_will_not_call_function(self):
        self.clock.schedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_once(self.callback_a, 0)
        self.clock.unschedule(self.callback_a)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.assertEqual(self.callback_a.call_count, 0)
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.assertEqual(self.callback_a.call_count, 0)

    def test_unschedule_will_not_fail_if_already_unscheduled(self):
        self.clock.schedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.clock.schedule_once(self.callback_a, 0)
        self.clock.unschedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.clock.schedule_interval(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.clock.unschedule(self.callback_a)
        self.clock.schedule_interval_soft(self.callback_a, 1)
        self.clock.unschedule(self.callback_a)
        self.clock.unschedule(self.callback_a)

    def test_call_sched_return_True_if_called_functions(self):
        self.clock.schedule(self.callback_a)
        self.assertTrue(self.clock.call_scheduled_functions(0))

    @unittest.skip('Requires changes to the clock')
    def test_call_sched_return_True_if_called_functions_interval(self):
        self.clock.schedule_once(self.callback_a, 1)
        self.assertFalse(self.clock.call_scheduled_functions(0))
        self.clock.set_time(1)
        self.assertTrue(self.clock.call_scheduled_functions(0))

    def test_call_sched_return_False_if_no_called_functions(self):
        self.assertFalse(self.clock.call_scheduled_functions(0))

    def test_tick_return_last_delta(self):
        self.assertEqual(self.clock.tick(), 0)
        self.time = 1
        self.assertEqual(self.clock.tick(), 1)
        self.time = 3
        self.assertEqual(self.clock.tick(), 2)

    @unittest.skip('Requires changes to the clock')
    def test_get_sleep_time_None_if_no_items(self):
        self.assertIsNone(self.clock.get_sleep_time())

    @unittest.skip('Requires changes to the clock')
    def test_get_sleep_time_can_sleep(self):
        self.clock.schedule_once(self.callback_a, 3)
        self.clock.schedule_once(self.callback_b, 1)
        self.clock.schedule_once(self.callback_c, 6)
        self.clock.schedule_once(self.callback_d, 7)
        self.assertEqual(self.clock.get_sleep_time(), 1)
        self.advance_clock()
        self.assertEqual(self.clock.get_sleep_time(), 2)
        self.advance_clock(2)
        self.assertEqual(self.clock.get_sleep_time(), 3)
        self.advance_clock(3)
        self.assertEqual(self.clock.get_sleep_time(), 1)

    @unittest.skip('Requires changes to the clock')
    def test_get_sleep_time_cannot_sleep(self):
        self.clock.schedule(self.callback_a)
        self.clock.schedule_once(self.callback_b, 1)
        self.assertEqual(self.clock.get_sleep_time(), 0)

    @unittest.skip
    def test_schedule_item_during_tick(self):
        def replicating_event(dt):
            self.clock.schedule(replicating_event)
            counter()

        counter = mock.Mock()
        self.clock.schedule(replicating_event)

        # one tick for the original event
        self.clock.tick()
        self.assertEqual(counter.call_count, 1)

        # requires access to private member
        self.assertEqual(len(self.clock._schedule_items), 2)

        # one tick from original, then two for new
        # now event queue should have two items as well
        self.clock.tick()
        self.assertEqual(counter.call_count, 3)

        # requires access to private member
        self.assertEqual(len(self.clock._schedule_items), 4)

    def test_unschedule_interval_item_during_tick(self):
        def suicidal_event(dt):
            counter()
            self.clock.unschedule(suicidal_event)

        counter = mock.Mock()
        self.clock.schedule_interval(suicidal_event, 1)
        self.advance_clock()
        self.assertEqual(counter.call_count, 1)

    @unittest.skip
    def test_schedule_interval_item_during_tick(self):
        def replicating_event(dt):
            self.clock.schedule_interval(replicating_event, 1)
            counter()

        counter = mock.Mock()
        self.clock.schedule_interval(replicating_event, 1)

        # advance time for the original event
        self.advance_clock()
        self.assertEqual(counter.call_count, 1)

        # requires access to private member
        self.assertEqual(len(self.clock._schedule_interval_items), 2)

        # one tick from original, then two for new
        # now event queue should have two items as well
        self.advance_clock()
        self.assertEqual(counter.call_count, 3)

        # requires access to private member
        self.assertEqual(len(self.clock._schedule_interval_items), 4)

    def test_scheduler_integrity(self):
        """most tests in this suite do not care about which order
        scheduled items are executed.  this test will verify that
        the order things are executed is correct.
        """
        expected_order = [self.callback_a, self.callback_b,
                          self.callback_c, self.callback_d]

        # schedule backwards to verify that they are scheduled correctly,
        # even if scheduled out-of-order.
        for delay, func in reversed(list(enumerate(expected_order, start=1))):
            self.clock.schedule_once(func, delay)

        for index, func in enumerate(expected_order, start=1):
            self.advance_clock()
            self.assertTrue(func.called)
            self.assertFalse(any(i.called for i in expected_order[index:]))

    def test_slow_clock(self):
        """pyglet's clock will not make up for lost time.  in this case, the
        interval scheduled for callback_[bcd] is 1, and 2 seconds have passed.
        since pyglet won't make up for lost time, they are only called once.
        """
        self.clock.schedule(self.callback_a)
        self.clock.schedule_once(self.callback_b, 1)
        self.clock.schedule_interval(self.callback_c, 1)
        self.clock.schedule_interval_soft(self.callback_d, 1)

        # simulate a slow clock
        self.time = 2
        self.clock.tick()

        self.assertEqual(self.callback_a.call_count, 1)
        self.assertEqual(self.callback_b.call_count, 1)
        self.assertEqual(self.callback_c.call_count, 1)
        self.assertEqual(self.callback_d.call_count, 1)

    def test_slow_clock_reschedules(self):
        """pyglet's clock will not make up for lost time.  in this case, the
        interval scheduled for callback_[bcd] is 1, and 2 seconds have passed.
        since pyglet won't make up for lost time (call events that missed their
        execution time), they are only called once.  this test verifies that
        missed events are rescheduled and executed later
        """
        self.clock.schedule(self.callback_a)
        self.clock.schedule_once(self.callback_b, 1)
        self.clock.schedule_interval(self.callback_c, 1)
        self.clock.schedule_interval_soft(self.callback_d, 1)

        # simulate slow clock
        self.time = 2
        self.clock.tick()

        # simulate a proper clock (advance clock time by one)
        frames = self.advance_clock()

        # make sure our clock is at 3 seconds
        self.assertEqual(self.time, 3)

        # the +1 is the call during the slow clock period
        self.assertEqual(self.callback_a.call_count, frames + 1)

        # only scheduled to happen once
        self.assertEqual(self.callback_b.call_count, 1)

        # 2 because they 'missed' a call when the clock lagged
        # with a good clock, this would be 3
        self.assertEqual(self.callback_c.call_count, 2)
        self.assertEqual(self.callback_d.call_count, 2)

    @unittest.skip('Requires changes to the clock')
    def test_get_interval(self):
        self.assertEqual(self.clock.get_interval(), 0)
        self.advance_clock(100)
        self.assertEqual(round(self.clock.get_interval(), 10), self.interval)

    def test_soft_scheduling_stress_test(self):
        """test that the soft scheduler is able to correctly soft-schedule
        several overlapping events.
        this test delves into implementation of the clock, and may break
        """
        # this value represents evenly scheduled items between 0 & 1
        # and what is produced by the correct soft-scheduler
        expected = [0.0625, 0.125, 0.1875, 0.25, 0.3125, 0.375, 0.4375, 0.5,
                    0.5625, 0.625, 0.6875, 0.75, 0.8125, 0.875, 0.9375, 1]

        for i in range(16):
            self.clock.schedule_interval_soft(None, 1)

        # sort the clock items
        items = sorted(i.next_ts for i in self.clock._schedule_interval_items)

        self.assertEqual(items, expected)



"""Wrapper for include/libavformat/avformat.h
"""
from ctypes import c_int, c_int64
from ctypes import c_uint8, c_uint, c_double, c_ubyte, c_size_t, c_char, c_char_p
from ctypes import c_void_p, POINTER, CFUNCTYPE, Structure, Union

import pyglet.lib
from pyglet.util import debug_print
from . import compat
from . import libavcodec
from . import libavutil

_debug = debug_print('debug_media')

avformat = pyglet.lib.load_library(
    'avformat',
    win32=('avformat-61', 'avformat-60', 'avformat-59', 'avformat-58'),
    darwin=('avformat.61', 'avformat.60', 'avformat.59', 'avformat.58')
)

avformat.avformat_version.restype = c_int

avformat_version = avformat.avformat_version() >> 16

compat.set_version('avformat', avformat_version)

AVSEEK_FLAG_BACKWARD = 1  # ///< seek backward
AVSEEK_FLAG_BYTE = 2  # ///< seeking based on position in bytes
AVSEEK_FLAG_ANY = 4  # ///< seek to any frame, even non-keyframes
AVSEEK_FLAG_FRAME = 8  # ///< seeking based on frame number
AVSEEK_SIZE = 0x10000
AVFMT_FLAG_CUSTOM_IO = 0x0080

MAX_REORDER_DELAY = 16


class AVPacketList(Structure):
    pass


class AVInputFormat(Structure):
    _fields_ = [
        ('name', c_char_p)
    ]


class AVOutputFormat(Structure):
    pass


class AVIOContext(Structure):
    pass


class AVIndexEntry(Structure):
    pass


class AVStreamInfo(Structure):
    _fields_ = [
        ('last_dts', c_int64),
        ('duration_gcd', c_int64),
        ('duration_count', c_int),
        ('rfps_duration_sum', c_int64),
        ('duration_error', POINTER(c_double * 2 * (30 * 12 + 30 + 3 + 6))),
        ('codec_info_duration', c_int64),
        ('codec_info_duration_fields', c_int64),
        ('frame_delay_evidence', c_int),
        ('found_decoder', c_int),
        ('last_duration', c_int64),
        ('fps_first_dts', c_int64),
        ('fps_first_dts_idx', c_int),
        ('fps_last_dts', c_int64),
        ('fps_last_dts_idx', c_int),
    ]


class AVProbeData(Structure):
    _fields_ = [
        ('filename', c_char_p),
        ('buf', POINTER(c_ubyte)),
        ('buf_size', c_int),
        ('mime_type', c_char_p)
    ]


class FFFrac(Structure):
    pass


class AVStreamInternal(Structure):
    pass


class AVFrac(Structure):
    _fields_ = [
        ('val', c_int64),
        ('num', c_int64),
        ('den', c_int64),
    ]


AVCodecContext = libavcodec.AVCodecContext
AVPacketSideData = libavcodec.AVPacketSideData
AVPacket = libavcodec.AVPacket
AVCodecParserContext = libavcodec.AVCodecParserContext
AVCodecParameters = libavcodec.AVCodecParameters
AVRational = libavutil.AVRational
AVDictionary = libavutil.AVDictionary
AVFrame = libavutil.AVFrame
AVClass = libavutil.AVClass
AVCodec = libavcodec.AVCodec



class AVStream(Structure):
    pass

AVStream_Fields = [
        ('av_class', POINTER(AVClass)),
        ('index', c_int),
        ('id', c_int),
        ('codec', POINTER(AVCodecContext)),  # Deprecated. Removed in 59.
        ('priv_data', c_void_p),
        ('time_base', AVRational),
        ('start_time', c_int64),
        ('duration', c_int64),
        ('nb_frames', c_int64),
        ('disposition', c_int),
        ('discard', c_int),
        ('sample_aspect_ratio', AVRational),
        ('metadata', POINTER(AVDictionary)),
        ('avg_frame_rate', AVRational),
        ('attached_pic', AVPacket),
        ('side_data', POINTER(AVPacketSideData)),
        ('nb_side_data', c_int),
        ('event_flags', c_int),
        ('r_frame_rate', AVRational),
        ('recommended_encoder_configuration', c_char_p),  # Deprecated. Removed in 59.
        ('codecpar', POINTER(AVCodecParameters)),
        ('info', POINTER(AVStreamInfo)),  # Deprecated. Removed in 59.
        ('pts_wrap_bits', c_int),
    ]

compat.add_version_changes('avformat', 58, AVStream, AVStream_Fields, removals=('av_class',))

compat.add_version_changes('avformat', 59, AVStream, AVStream_Fields,
                           removals=('av_class', 'codec', 'recommended_encoder_configuration', 'info'))

for compat_ver in (60, 61):
    compat.add_version_changes('avformat', compat_ver, AVStream, AVStream_Fields,
                               removals=('codec', 'recommended_encoder_configuration', 'info'),
                               repositions=(compat.Reposition("codecpar", "id"),))


class AVProgram(Structure):
    pass


class AVChapter(Structure):
    pass


class AVFormatInternal(Structure):
    pass


class AVIOInterruptCB(Structure):
    _fields_ = [
        ('callback', CFUNCTYPE(c_int, c_void_p)),
        ('opaque', c_void_p)
    ]

class AVIAMFAudioElement(Structure):
    ...

class AVIAMFMixPresentation(Structure):
    ...

class _AvStreamGroupTileGridOffsets(Structure):
    _fields_ = [
        ('idx', c_uint),
        ('horizontal', c_int),
        ('vertical', c_int),
    ]

class AVStreamGroupTileGrid(Structure):
    _fields_ = [
        ('av_class', POINTER(AVClass)),
        ('nb_tiles', c_uint),
        ('coded_width', c_int),
        ('coded_height', c_int),
        ('offsets', POINTER(_AvStreamGroupTileGridOffsets)),
        ('background', c_uint8 * 4),
        ('horizontal_offset', c_int),
        ('vertical_offset', c_int),
        ('width', c_int),
        ('height', c_int),
    ]

class AVStreamGroupLCEVC(Structure):
    _fields_ = [
        ('av_class', POINTER(AVClass)),
        ('lcevc_index', c_int),
        ('width', c_int),
        ('height', c_int)
    ]

class _AVStreamGroupParams(Union):
    _fields_ = [
        ('iamf_audio_element', POINTER(AVIAMFAudioElement)),
        ('iamf_mix_presentation', POINTER(AVIAMFMixPresentation)),
        ('tile_grid', POINTER(AVStreamGroupTileGrid)),
        ('lcevc', POINTER(AVStreamGroupLCEVC))
    ]

class AVStreamGroup(Structure):
    _fields_ = [
        ('av_class', POINTER(AVClass)),
        ('priv_data', c_void_p),
        ('index', c_uint),
        ('id', c_int64),
        ('type', c_int),
        ('params', _AVStreamGroupParams),
        ('metadata', POINTER(AVDictionary)),
        ('nb_streams', c_uint),
        ('streams', POINTER(POINTER(AVStream))),
        ('disposition', c_int),
    ]


class AVFormatContext(Structure):
    pass


if avformat_version >= 61:
    AVFormatContext_Fields = [
        ('av_class', POINTER(AVClass)),
        ('iformat', POINTER(AVInputFormat)),
        ('oformat', POINTER(AVOutputFormat)),
        ('priv_data', c_void_p),
        ('pb', POINTER(AVIOContext)),
        ('ctx_flags', c_int),
        ('nb_streams', c_uint),
        ('streams', POINTER(POINTER(AVStream))),
        ('nb_stream_groups', c_uint),
        ('stream_groups', POINTER(POINTER(AVStreamGroup))),
        ('nb_chapters', c_uint),  # Moved after streams in 7.x
        ('chapters', POINTER(POINTER(AVChapter))),  # Moved after streams in 7.x
        ('url', c_char_p),
        ('start_time', c_int64),
        ('duration', c_int64),
        ('bit_rate', c_int64),
        ('packet_size', c_uint),
        ('max_delay', c_int),
        ('flags', c_int),
        ('probesize', c_int64),
        ('max_analyze_duration', c_int64),
        ('key', POINTER(c_uint8)),
        ('keylen', c_int),
        ('nb_programs', c_uint),
        ('programs', POINTER(POINTER(AVProgram))),
        ('video_codec_id', c_int),
        ('audio_codec_id', c_int),
        ('subtitle_codec_id', c_int),
        ('data_codec_id', c_int),
        ('metadata', POINTER(AVDictionary)),
        ('start_time_realtime', c_int64),
        ('fps_probe_size', c_int),
        ('error_recognition', c_int),
        ('interrupt_callback', AVIOInterruptCB),
        ('debug', c_int),
        ('max_streams', c_int),
        ('max_index_size', c_uint),
        ('max_picture_buffer', c_uint),
        ('max_interleave_delta', c_int64),
        ('max_ts_probe', c_int),
        ('max_chunk_duration', c_int),
        ('max_chunk_size', c_int),
        ('max_probe_packets', c_int),
        ('strict_std_compliance', c_int),
        ('event_flags', c_int),
        ('avoid_negative_ts', c_int),
        ('audio_preload', c_int),
        ('use_wallclock_as_timestamps', c_int),
        ('skip_estimate_duration_from_pts', c_int),  # Added in 59.
        ('avio_flags', c_int),
        ('duration_estimation_method', c_uint),
        ('skip_initial_bytes', c_int64),
        ('correct_ts_overflow', c_uint),
        ('seek2any', c_int),
        ('flush_packets', c_int),
        ('probe_score', c_int),
        ('format_probesize', c_int),
        ('codec_whitelist', c_char_p),
        ('format_whitelist', c_char_p),
        ('protocol_whitelist', c_char_p),
        ('protocol_blacklist', c_char_p),
        ('io_repositioned', c_int),
        ('video_codec', POINTER(AVCodec)),
        ('audio_codec', POINTER(AVCodec)),
        ('subtitle_codec', POINTER(AVCodec)),
        ('data_codec', POINTER(AVCodec)),
        ('metadata_header_padding', c_int),
        ('opaque', c_void_p),
        ('control_message_cb', CFUNCTYPE(c_int,
                                         POINTER(AVFormatContext), c_int, c_void_p,
                                         c_size_t)),
        ('output_ts_offset', c_int64),
        ('dump_separator', POINTER(c_uint8)),

        ('io_open', CFUNCTYPE(c_int,
                              POINTER(AVFormatContext),
                              POINTER(POINTER(AVIOContext)),
                              c_char_p, c_int,
                              POINTER(POINTER(AVDictionary)))),

        ('io_close2', CFUNCTYPE(c_int, POINTER(AVFormatContext), POINTER(AVIOContext)))  # Added in 59.
    ]

    compat.add_version_changes('avformat', 61, AVFormatContext, AVFormatContext_Fields, removals=None)


else:
    AVFormatContext_Fields = [
        ('av_class', POINTER(AVClass)),
        ('iformat', POINTER(AVInputFormat)),
        ('oformat', POINTER(AVOutputFormat)),
        ('priv_data', c_void_p),
        ('pb', POINTER(AVIOContext)),
        ('ctx_flags', c_int),
        ('nb_streams', c_uint),
        ('streams', POINTER(POINTER(AVStream))),
        ('filename', c_char * 1024),  # Deprecated. Removed in 59
        ('url', c_char_p),
        ('start_time', c_int64),
        ('duration', c_int64),
        ('bit_rate', c_int64),
        ('packet_size', c_uint),
        ('max_delay', c_int),
        ('flags', c_int),
        ('probesize', c_int64),
        ('max_analyze_duration', c_int64),
        ('key', POINTER(c_uint8)),
        ('keylen', c_int),
        ('nb_programs', c_uint),
        ('programs', POINTER(POINTER(AVProgram))),
        ('video_codec_id', c_int),
        ('audio_codec_id', c_int),
        ('subtitle_codec_id', c_int),
        ('max_index_size', c_uint),
        ('max_picture_buffer', c_uint),
        ('nb_chapters', c_uint),  # Moved after streams in 7.x
        ('chapters', POINTER(POINTER(AVChapter))),  # Moved after streams in 7.x
        ('metadata', POINTER(AVDictionary)),
        ('start_time_realtime', c_int64),
        ('fps_probe_size', c_int),
        ('error_recognition', c_int),
        ('interrupt_callback', AVIOInterruptCB),
        ('debug', c_int),
        ('max_interleave_delta', c_int64),
        ('strict_std_compliance', c_int),
        ('event_flags', c_int),
        ('max_ts_probe', c_int),
        ('avoid_negative_ts', c_int),
        ('ts_id', c_int),
        ('audio_preload', c_int),
        ('max_chunk_duration', c_int),
        ('max_chunk_size', c_int),
        ('use_wallclock_as_timestamps', c_int),
        ('avio_flags', c_int),
        ('duration_estimation_method', c_uint),
        ('skip_initial_bytes', c_int64),
        ('correct_ts_overflow', c_uint),
        ('seek2any', c_int),
        ('flush_packets', c_int),
        ('probe_score', c_int),
        ('format_probesize', c_int),
        ('codec_whitelist', c_char_p),
        ('format_whitelist', c_char_p),
        ('internal', POINTER(AVFormatInternal)),  # Deprecated. Removed in 59
        ('io_repositioned', c_int),
        ('video_codec', POINTER(AVCodec)),
        ('audio_codec', POINTER(AVCodec)),
        ('subtitle_codec', POINTER(AVCodec)),
        ('data_codec', POINTER(AVCodec)),
        ('metadata_header_padding', c_int),
        ('opaque', c_void_p),
        ('control_message_cb', CFUNCTYPE(c_int,
                                         POINTER(AVFormatContext), c_int, c_void_p,
                                         c_size_t)),
        ('output_ts_offset', c_int64),
        ('dump_separator', POINTER(c_uint8)),
        ('data_codec_id', c_int),
        ('protocol_whitelist', c_char_p),
        ('io_open', CFUNCTYPE(c_int,
                              POINTER(AVFormatContext),
                              POINTER(POINTER(AVIOContext)),
                              c_char_p, c_int,
                              POINTER(POINTER(AVDictionary)))),
        ('io_close', CFUNCTYPE(None,
                               POINTER(AVFormatContext), POINTER(AVIOContext))),
        ('protocol_blacklist', c_char_p),
        ('max_streams', c_int),
        ('skip_estimate_duration_from_pts', c_int),  # Added in 59.
        ('max_probe_packets', c_int), # Added in 59.
        ('io_close2', CFUNCTYPE(c_int, POINTER(AVFormatContext), POINTER(AVIOContext))) # Added in 59.
    ]

    compat.add_version_changes('avformat', 58, AVFormatContext, AVFormatContext_Fields,
                               removals=('skip_estimate_duration_from_pts', 'max_probe_packets', 'io_close2'))

    for compat_ver in (59, 60):
        compat.add_version_changes('avformat', compat_ver, AVFormatContext, AVFormatContext_Fields,
                                   removals=('filename', 'internal'))

avformat.av_find_input_format.restype = c_int
avformat.av_find_input_format.argtypes = [c_int]
avformat.avformat_open_input.restype = c_int
avformat.avformat_open_input.argtypes = [
    POINTER(POINTER(AVFormatContext)),
    c_char_p,
    POINTER(AVInputFormat),
    POINTER(POINTER(AVDictionary))]
avformat.avformat_find_stream_info.restype = c_int
avformat.avformat_find_stream_info.argtypes = [
    POINTER(AVFormatContext),
    POINTER(POINTER(AVDictionary))]
avformat.avformat_close_input.restype = None
avformat.avformat_close_input.argtypes = [
    POINTER(POINTER(AVFormatContext))]
avformat.av_read_frame.restype = c_int
avformat.av_read_frame.argtypes = [POINTER(AVFormatContext),
                                   POINTER(AVPacket)]
avformat.av_seek_frame.restype = c_int
avformat.av_seek_frame.argtypes = [POINTER(AVFormatContext),
                                   c_int, c_int64, c_int]
avformat.avformat_seek_file.restype = c_int
avformat.avformat_seek_file.argtypes = [POINTER(AVFormatContext),
                                        c_int, c_int64, c_int64, c_int64, c_int]
avformat.av_guess_frame_rate.restype = AVRational
avformat.av_guess_frame_rate.argtypes = [POINTER(AVFormatContext),
                                         POINTER(AVStream), POINTER(AVFrame)]

ffmpeg_read_func = CFUNCTYPE(c_int, c_void_p, POINTER(c_char), c_int)
ffmpeg_write_func = CFUNCTYPE(c_int, c_void_p, POINTER(c_char), c_int)
ffmpeg_seek_func = CFUNCTYPE(c_int64, c_void_p, c_int64, c_int)

avformat.avio_alloc_context.restype = POINTER(AVIOContext)
avformat.avio_alloc_context.argtypes = [c_char_p, c_int, c_int, c_void_p, ffmpeg_read_func, c_void_p, ffmpeg_seek_func]

avformat.avformat_alloc_context.restype = POINTER(AVFormatContext)
avformat.avformat_alloc_context.argtypes = []

avformat.avformat_free_context.restype = c_void_p
avformat.avformat_free_context.argtypes = [POINTER(AVFormatContext)]

__all__ = [
    'avformat',
    'AVSEEK_FLAG_BACKWARD',
    'AVSEEK_FLAG_BYTE',
    'AVSEEK_FLAG_ANY',
    'AVSEEK_FLAG_FRAME',
    'AVFormatContext',
    'AVCodecContext',
    'avformat_version',
]



"""WIP."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyglet.gui.widgets import WidgetBase
    from pyglet.window import BaseWindow


class Frame:
    """The base Frame object, implementing a 2D spatial hash.

    A `Frame` provides an efficient way to handle dispatching
    keyboard and mouse events to Widgets. This is done by
    implementing a 2D spatial hash. Only Widgets that are in the
    vicinity of the mouse pointer will be passed Window events,
    which can greatly improve efficiency when a large quantity
    of Widgets are in use.
    """

    def __init__(self, window: BaseWindow, enable: bool = True, cell_size: int = 64, order: int = 0) -> None:
        """Create an instance of a Frame.

        Args:
            window:
                The SpatialHash will receive events from this Window.
                Appropriate events will be passed on to all added Widgets.
            enable:
                Whether to enable frame.
            cell_size:
                The cell ("bucket") size for each cell in the hash.
                Widgets may span multiple cells.
            order:
                Widgets use internal ordered Groups for draw sorting.
                This is the base value for these Groups.
        """
        self._window = window
        self._enable = enable
        self._cell_size = cell_size
        self._cells = {}
        self._active_widgets = set()
        self._order = order
        self._mouse_pos = 0, 0
        if self._enable:
            self._window.push_handlers(self)

    def _hash(self, x: float, y: float) -> tuple[int, int]:
        """Normalize position to cell."""
        return int(x / self._cell_size), int(y / self._cell_size)

    def _on_reposition_handler(self, widget):
        self.remove_widget(widget)
        self.add_widget(widget)

    @property
    def enable(self):
        """Whether to enable frame.

        :type: bool
        """
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = bool(value)
        if self._enable:
            self._window.push_handlers(self)
        else:
            self._window.remove_handlers(self)

    def add_widget(self, widget: WidgetBase) -> None:
        """Add a Widget to the spatial hash."""
        min_vec, max_vec = self._hash(*widget.aabb[0:2]), self._hash(*widget.aabb[2:4])
        for i in range(min_vec[0], max_vec[0] + 1):
            for j in range(min_vec[1], max_vec[1] + 1):
                self._cells.setdefault((i, j), set()).add(widget)
        widget.update_groups(self._order)
        widget.set_handler("on_reposition", self._on_reposition_handler)

    def remove_widget(self, widget: WidgetBase) -> None:
        """Remove a Widget from the spatial hash."""
        min_vec, max_vec = self._hash(*widget.aabb[0:2]), self._hash(*widget.aabb[2:4])
        for i in range(min_vec[0], max_vec[0] + 1):
            for j in range(min_vec[1], max_vec[1] + 1):
                self._cells.get((i, j)).remove(widget)

    # Handlers

    def on_key_press(self, symbol: int, modifiers: int) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(*self._mouse_pos), set()):
            widget.on_key_press(symbol, modifiers)

    def on_key_release(self, symbol: int, modifiers: int) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(*self._mouse_pos), set()):
            widget.on_key_release(symbol, modifiers)

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(x, y), set()):
            widget.on_mouse_press(x, y, buttons, modifiers)
            self._active_widgets.add(widget)

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        """Pass the event to any widgets that are currently active."""
        for widget in self._active_widgets:
            widget.on_mouse_release(x, y, buttons, modifiers)
        self._active_widgets.clear()

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        """Pass the event to any widgets that are currently active."""
        for widget in self._active_widgets:
            widget.on_mouse_drag(x, y, dx, dy, buttons, modifiers)
        self._mouse_pos = x, y

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(x, y), set()):
            widget.on_mouse_scroll(x, y, scroll_x, scroll_y)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        """Pass the event to any widgets within range of the mouse"""
        for cell in self._cells.values():
            for widget in cell:
                widget.on_mouse_motion(x, y, dx, dy)
        self._mouse_pos = x, y

    def on_text(self, text: str) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(*self._mouse_pos), set()):
            widget.on_text(text)

    def on_text_motion(self, motion: int) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(*self._mouse_pos), set()):
            widget.on_text_motion(motion)

    def on_text_motion_select(self, motion: int) -> None:
        """Pass the event to any widgets within range of the mouse."""
        for widget in self._cells.get(self._hash(*self._mouse_pos), set()):
            widget.on_text_motion_select(motion)


class MovableFrame(Frame):
    """A Frame that allows Widget repositioning.

    When a specified modifier key is held down, Widgets can be
    repositioned by dragging them. Examples of modifier keys are
    Ctrl, Alt, Shift. These are defined in the `pyglet.window.key`
    module, and start with `MOD_`. For example::

        from pyglet.window.key import MOD_CTRL

        frame = pyglet.gui.frame.MovableFrame(mywindow, modifier=MOD_CTRL)

    For more information, see the `pyglet.window.key` submodule
    API documentation.
    """

    def __init__(self, window: BaseWindow, enable: bool = True, order: int = 0, modifier: int = 0) -> None:
        """Create an instance of a MovableFrame.

        This is a similar to the standard Frame class, except that
        you can specify a modifier key. When this key is held down,
        Widgets can be re-positioned by drag-and-dropping.

        Args:
            window:
                The SpatialHash will receive events from this Window.
                Appropriate events will be passed on to all added Widgets.
            enable:
                Whether to enable frame.
            order:
                Widgets use internal ordered Groups for draw sorting.
                This is the base value for these Groups.
            modifier:
                A key modifier, such as `pyglet.window.key.MOD_CTRL`
        """
        super().__init__(window, enable=enable, order=order)
        self._modifier = modifier
        self._moving_widgets = set()

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if self._modifier & modifiers > 0:
            for widget in self._cells.get(self._hash(x, y), set()):
                if widget._check_hit(x, y):     # noqa: SLF001
                    self._moving_widgets.add(widget)
            for widget in self._moving_widgets:
                self.remove_widget(widget)
        else:
            super().on_mouse_press(x, y, buttons, modifiers)

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        for widget in self._moving_widgets:
            self.add_widget(widget)
        self._moving_widgets.clear()
        super().on_mouse_release(x, y, buttons, modifiers)

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        for widget in self._moving_widgets:
            wx, wy = widget.position
            widget.position = wx + dx, wy + dy
        super().on_mouse_drag(x, y, dx, dy, buttons, modifiers)



import pyglet


class PhysicalObject(pyglet.sprite.Sprite):
    """A sprite with physical properties such as velocity"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # In addition to position, we have velocity
        self.velocity_x, self.velocity_y = 0.0, 0.0

    def update(self, dt):
        """This method should be called every frame."""

        # Update position according to velocity and time
        self.x += self.velocity_x * dt
        self.y += self.velocity_y * dt

        # Wrap around the screen if necessary
        self.check_bounds()

    def check_bounds(self):
        """Use the classic Asteroids screen wrapping behavior"""
        min_x = -self.image.width / 2
        min_y = -self.image.height / 2
        max_x = 800 + self.image.width / 2
        max_y = 600 + self.image.height / 2
        if self.x < min_x:
            self.x = max_x
        elif self.x > max_x:
            self.x = min_x
        if self.y < min_y:
            self.y = max_y
        elif self.y > max_y:
            self.y = min_y



import pytest
from pyglet.math import Quaternion, Mat3, Mat4


def test_create():
    q = Quaternion()
    assert q.x == 0
    assert q.y == 0
    assert q.z == 0
    assert q.w == 1
    assert q.length() == 1


def test_from_mat3():
    with pytest.raises(NotImplementedError):
        Quaternion.from_mat3()


def test_from_mat4():
    with pytest.raises(NotImplementedError):
        Quaternion.from_mat4()


def test_to_mat3():
    # TODO: Needs improvement. Just checking identity matrix.
    q = Quaternion()
    assert q.to_mat3() == Mat3()


def test_to_mat4():
    # TODO: Needs improvement. Just checking identity matrix.
    q = Quaternion()
    assert q.to_mat4() == Mat4()



from .darwin_hid import get_devices
from .darwin_hid import get_joysticks
from .darwin_hid import get_apple_remote
from .darwin_hid import get_controllers
from .darwin_hid import DarwinControllerManager as ControllerManager


def get_tablets(display=None):
    import warnings
    warnings.warn("Tablets not yet supported on macOS.")
    return []



# Utilities to patch Pyglet to provide more precise event timing on Windows.
# HighTimerResolution is generally the most important of the bunch,
# since most 60+fps games in Windows will not work right without
# changing the timer resolution for the app

import pyglet
import ctypes

try:
    winmm = ctypes.windll.winmm
    k32 = ctypes.windll.kernel32
except AttributeError:
    raise NotImplementedError("This is not a Win32 compatible platform.")


class HighTimerResolution:
    """
    A context manager to adjust the main Win32 platform event loop.    
    
    Events inside the context manager will use the specified resolution of the
    multimedia timer to provide proper timings for 60fps+ framerates.    
    On exiting the context manager, the default resolution timer is restored.

    Most 60fps games can use a resolution of 4, but you may need to go as low
    as 1 to get the most consistent results.

    Example:

    with HighTimerResolution():
        pyglet.app.run()

    """

    def __init__(self, resolution=1):
        self.resolution = resolution

    def __enter__(self):
        winmm.timeBeginPeriod(self.resolution)

    def __exit__(self, *a):
        winmm.timeEndPeriod()


def pin_thread(pin_thread_mask=0x01):
    """
    Pin the main Pyglet app thread to one or more cores by way of a thread affinity mask.
    The default thread mask is 1.
    """
    k32.SetThreadAffinityMask(
        pyglet.app.platform_event_loop._event_thread, pin_thread_mask
    )


def pin_process():
    """
    Pins the entire Python interpreter process to the first CPU core.
    This may slightly improve performance due to affinity.
    """
    k32.SetProcessAffinityMask(k32.GetCurrentProcess(), 1)


def raise_priority(
    priority_class=0x080,
):
    """
    Raises the process priority to High. Other priorities can be provided.
    """
    k32.SetPriorityClass(-1, priority_class)



"""Precise framerate calculation function scheduling.

The :py:mod:`~pyglet.clock` module allows you to schedule functions
to run periodically, or for one-shot future execution. pyglet's default
event loop (:py:func:`~pyglet.app.run`) keeps an internal instance of
a :py:class:`~pyglet.clock.Clock`, which is ticked automatically.

.. note:: Some internal modules will schedule items on the clock. If you
          are using a custom event loop, always remember to `tick` the clock!

Scheduling
==========

You can schedule a function to be called every time the clock is ticked::

    def callback(dt):
        print(f"{dt} seconds since last callback")

    clock.schedule(callback)

The `schedule_interval` method causes a function to be called every "n"
seconds::

    clock.schedule_interval(callback, 0.5)   # called twice a second

The `schedule_once` method causes a function to be called once "n" seconds
in the future::

    clock.schedule_once(callback, 5)        # called in 5 seconds

All the `schedule` methods will pass on any additional args or keyword args
you specify to the callback function::

    def move(dt, velocity, sprite):
        sprite.position += dt * velocity

    clock.schedule(move, velocity=5.0, sprite=alien)

You can cancel a function scheduled with any of these methods using
`unschedule`::

    clock.unschedule(move)

Using multiple clocks
=====================

The clock functions are all relayed to an instance of
:py:class:`~pyglet.clock.Clock` which is initialised with the module.  You can
get this instance to use directly::

    clk = pyglet.clock.get_default()

You can also replace the default clock with your own:

    myclk = pyglet.clock.Clock()
    pyglet.clock.set_default(myclk)

Each clock maintains its own set of scheduled functions and frequency
measurement.  Each clock must be "ticked" separately.

Multiple and derived clocks potentially allow you to separate "game-time" and
"wall-time", or to synchronise your clock to an audio or video stream instead
of the system clock.
"""
from __future__ import annotations

import time as _time

from typing import Any, Callable

from heapq import heappop as _heappop
from heapq import heappush as _heappush
from heapq import heappushpop as _heappushpop
from operator import attrgetter as _attrgetter
from collections import deque as _deque


class _ScheduledItem:
    __slots__ = ['func', 'args', 'kwargs']

    def __init__(self, func: Callable, args: Any, kwargs: Any) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs


class _ScheduledIntervalItem:
    __slots__ = ['func', 'interval', 'last_ts', 'next_ts', 'args', 'kwargs']

    def __init__(self, func: Callable, interval: float, last_ts: float, next_ts: float, args: Any, kwargs: Any) -> None:
        self.func = func
        self.interval = interval
        self.last_ts = last_ts
        self.next_ts = next_ts
        self.args = args
        self.kwargs = kwargs

    def __lt__(self, other: _ScheduledIntervalItem) -> bool:
        return self.next_ts < other.next_ts


class Clock:

    # List of functions to call every tick.
    _schedule_items: list

    # List of schedule interval items kept in sort order.
    _schedule_interval_items: list

    # If True, a sleep(0) is inserted on every tick.
    _force_sleep: bool = False

    def __init__(self, time_function: Callable = _time.perf_counter) -> None:
        """Initialise a Clock, with optional custom time function.

        You can provide a custom time function to return the elapsed
        time of the application, in seconds. Defaults to ``time.perf_counter``,
        but can be replaced to allow for easy time dilation effects or game
        pausing.
        """
        self.time = time_function
        self.last_ts = None
        self.next_ts = self.time()

        # Used by self.get_frequency to show update frequency
        self.times: _deque = _deque()
        self.cumulative_time = 0.0
        self.window_size = 60

        self._schedule_items = []
        self._schedule_interval_items = []
        self._current_interval_item = None

    @staticmethod
    def sleep(microseconds: float) -> None:
        _time.sleep(microseconds * 1e-6)

    def update_time(self) -> float:
        """Get the elapsed time since the last call to `update_time`.

        This updates the clock's internal measure of time and returns
        the difference (in seconds) since the last time it was called.
        The first call of this method always returns 0.
        """
        ts = self.time()
        if self.last_ts is None:
            delta_t = 0.0
        else:
            delta_t = ts - self.last_ts
            self.times.appendleft(delta_t)
            if len(self.times) > self.window_size:
                self.cumulative_time -= self.times.pop()
        self.cumulative_time += delta_t
        self.last_ts = ts

        return delta_t

    def call_scheduled_functions(self, dt: float) -> bool:
        """Call scheduled functions that elapsed on the last `update_time`.

        This method is called automatically when the clock is ticked
        (see :py:meth:`~pyglet.clock.tick`), so you need not call it
        yourself in most cases.

        Args:
            dt:
                The elapsed time since the last update to pass to each
                scheduled function. This is *not* used to calculate which
                functions have elapsed.

        Returns: ``True`` if any functions were called, else ``False``.
        """
        now = self.last_ts or self.time()
        result = False  # flag indicates if any function was called

        # handle items scheduled for every tick
        if self._schedule_items:
            result = True
            # duplicate list in case event unschedules itself
            for item in list(self._schedule_items):
                item.func(dt, *item.args, **item.kwargs)

        # check the next scheduled item that is not called each tick
        # if it is scheduled in the future, then exit
        interval_items = self._schedule_interval_items
        try:
            if interval_items[0].next_ts > now:
                return result
        except IndexError:
            # The interval_items list is empty
            return result

        # NOTE: there is no special handling required to manage things
        #       that are scheduled during this loop, due to the heap
        self._current_interval_item = item = None
        get_soft_next_ts = self._get_soft_next_ts
        while interval_items:

            # the scheduler will hold onto a reference to an item in
            # case it needs to be rescheduled.  it is more efficient
            # to push and pop the heap at once rather than two operations
            if item is None:
                item = _heappop(interval_items)
            else:
                item = _heappushpop(interval_items, item)

            # a scheduled function may try to unschedule itself,
            # so we need to keep a reference to the current
            # item no longer on heap to be able to check
            self._current_interval_item = item

            # if next item is scheduled in the future then break
            if item.next_ts > now:
                break

            # execute the callback
            item.func(now - item.last_ts, *item.args, **item.kwargs)

            if item.interval:

                # Try to keep timing regular, even if overslept this time;
                # but don't schedule in the past (which could lead to
                # infinitely-worsening error).
                item.next_ts = item.last_ts + item.interval
                item.last_ts = now

                # test the schedule for the next execution
                if item.next_ts <= now:
                    # the scheduled time of this item has already
                    # passed, so it must be rescheduled
                    if now - item.next_ts < 0.05:
                        # missed execution time by 'reasonable' amount, so
                        # reschedule at normal interval
                        item.next_ts = now + item.interval
                    else:
                        # missed by significant amount, now many events have
                        # likely missed execution. do a soft re-schedule to
                        # avoid lumping many events together.
                        # in this case, the next dt will not be accurate
                        item.next_ts = get_soft_next_ts(now, item.interval)
                        item.last_ts = item.next_ts - item.interval
            else:
                # not an interval, so this item will not be rescheduled
                self._current_interval_item = item = None

        if item is not None:
            _heappush(interval_items, item)

        return True

    def tick(self, poll: bool = False) -> float:
        """Signify that one frame has passed.

        This will call any scheduled functions that have elapsed,
        and returns the number of seconds since the last time this
        method has been called. The first time this method is called,
        0 is returned.

        Args:
            poll:
                If True, the function will call any scheduled functions
                but will not sleep or busy-wait for any reason.  Recommended
                for advanced applications managing their own sleep timers
                only.
        """
        if not poll and self._force_sleep:
            self.sleep(0)

        delta_t = self.update_time()
        self.call_scheduled_functions(delta_t)
        return delta_t

    def get_sleep_time(self, sleep_idle: bool) -> float | None:
        """Get the time until the next item is scheduled, if any.

        Applications can choose to continue receiving updates at the
        maximum framerate during idle time (when no functions are scheduled),
        or they can sleep through their idle time and allow the CPU to
        switch to other processes or run in low-power mode.

        If ``sleep_idle`` is ``True`` the latter behaviour is selected, and
        ``None`` will be returned if there are no scheduled items.

        Otherwise, if ``sleep_idle`` is ``False``, or if any scheduled items
        exist, a value of 0 is returned.

        Args:
            sleep_idle:
                If True, the application intends to sleep through its idle
                time; otherwise it will continue ticking at the maximum
                frame rate allowed.
        """
        if self._schedule_items or not sleep_idle:
            return 0.0

        if self._schedule_interval_items:
            return max(self._schedule_interval_items[0].next_ts - self.time(), 0.0)

        return None

    def get_frequency(self) -> float:
        """Get the average clock update frequency of recent history.

        The result is the average of a sliding window of the last "n" updates,
        where "n" is some number designed to cover approximately 1 second.
        This is the clock frequency, **not** the Window redraw rate (fps).
        """
        if not self.cumulative_time:
            return 0
        return len(self.times) / self.cumulative_time

    def _get_nearest_ts(self) -> float:
        """Get the nearest timestamp.

        Schedule from now, unless now is sufficiently close to last_ts, in
        which case use last_ts.  This clusters together scheduled items that
        probably want to be scheduled together.
        """
        last_ts = self.last_ts or self.next_ts
        ts = self.time()
        if ts - last_ts > 0.2:
            return ts
        return last_ts

    def _get_soft_next_ts(self, last_ts: float, interval: float) -> float:

        def taken(ts: float, e: float) -> bool:
            """Check if `ts` has already got an item scheduled nearby."""
            # TODO this function is slow and called very often.
            # Optimise it, maybe?
            for item in self._schedule_interval_items:
                if abs(item.next_ts - ts) <= e:
                    return True
                elif item.next_ts > ts + e:
                    return False

            return False

        # sorted list is required to produce expected results
        # taken() will iterate through the heap, expecting it to be sorted
        # and will not always catch the smallest value, so sort here.
        # do not remove the sort key...it is faster than relaying comparisons
        # NOTE: do not rewrite as popping from heap, as that is super slow!
        self._schedule_interval_items.sort(key=_attrgetter('next_ts'))

        # Binary division over interval:
        #
        # 0                          interval
        # |--------------------------|
        #   5  3   6   2   7  4  8   1          Order of search
        #
        # i.e., first scheduled at interval,
        #       then at            interval/2
        #       then at            interval/4
        #       then at            interval*3/4
        #       then at            ...
        #
        # Schedule is hopefully then evenly distributed for any interval,
        # and any number of scheduled functions.

        next_ts = last_ts + interval
        if not taken(next_ts, interval / 4):
            return next_ts

        dt = interval
        divs = 1
        while True:
            next_ts = last_ts
            for _ in range(divs - 1):
                next_ts += dt
                if not taken(next_ts, dt / 4):
                    return next_ts
            dt /= 2
            divs *= 2

            # Avoid infinite loop in pathological case
            if divs > 16:
                return next_ts

    def schedule(self, func: Callable, *args: Any, **kwargs: Any) -> None:
        """Schedule a function to be called every tick.

        The scheduled function should have a prototype that includes ``dt``
        as the first argument, which gives the elapsed time in seconds since
        the last clock tick. Any additional args or kwargs given to this
        method are passed on to the callback::

            def callback(dt, *args, **kwargs):
                pass


        .. note:: Functions scheduled using this method will be called
                  every tick by the default pyglet event loop, which can
                  lead to high CPU usage. It is usually better to use
                  :py:meth:`~pyglet.clock.schedule_interval` unless
                  this is desired.
        """
        item = _ScheduledItem(func, args, kwargs)
        self._schedule_items.append(item)

    def schedule_once(self, func: Callable, delay: float, *args: Any, **kwargs: Any) -> None:
        """Schedule a function to be called once after ``delay`` seconds.

        The callback function prototype is the same as for
        :py:meth:`~pyglet.clock.Clock.schedule`.
        """
        last_ts = self._get_nearest_ts()
        next_ts = last_ts + delay
        item = _ScheduledIntervalItem(func, 0, last_ts, next_ts, args, kwargs)
        _heappush(self._schedule_interval_items, item)

    def schedule_interval(self, func: Callable, interval: float, *args: Any, **kwargs: Any) -> None:
        """Schedule a function to be called every ``interval`` seconds.

        To schedule a function to be called at 60Hz (60fps), you would use ``1/60``
        for the interval, and so on. If pyglet is unable to call the function on
        time, the schedule will be skipped (not accumulated). This can occur if the
        main thread is overloaded, or other hard blocking calls taking place.

        The callback function prototype is the same as for
        :py:meth:`~pyglet.clock.Clock.schedule`.

        .. note:: Specifying an interval of ``0`` will prevent the function from
                  being called again. If you want to schedule a function to be called
                  as often as possible, see :py:meth:`~pyglet.clock.Clock.schedule`.
        """
        last_ts = self._get_nearest_ts()
        next_ts = last_ts + interval
        item = _ScheduledIntervalItem(func, interval, last_ts, next_ts, args, kwargs)
        _heappush(self._schedule_interval_items, item)

    def schedule_interval_for_duration(self, func: Callable, interval: float,
                                       duration: float, *args: Any, **kwargs: Any) -> None:
        """Temporarily schedule a function to be called every ``interval`` seconds.

        This method will schedule a function to be called every ``interval``
        seconds (see  :py:meth:`~pyglet.clock.Clock.schedule_interval`), but
        will automatically unschedule it after ``duration`` seconds.

        The callback function prototype is the same as for
        :py:meth:`~pyglet.clock.Clock.schedule`.

        :Args:
            func:
                The function to call when the timer lapses.
            interval:
                The number of seconds to wait between each call.
            duration:
                The number of seconds for which the function is scheduled.
        """
        # NOTE: unschedule wrapper that takes `dt` argument
        def _unschedule(_dt: float, _func: Callable) -> None:
            self.unschedule(_func)

        self.schedule_interval(func, interval, *args, **kwargs)
        self.schedule_once(_unschedule, duration, func)

    def schedule_interval_soft(self, func: Callable, interval: float, *args: Any, **kwargs: Any) -> None:
        """Schedule a function to be called approximately every ``interval`` seconds.

        This method is similar to :py:meth:`~pyglet.clock.Clock.schedule_interval`,
        except that the clock will move the interval out of phase with other
        scheduled functions in order to distribute CPU load more evenly.

        This is useful for functions that need to be called regularly,
        but not relative to the initial start time.  :py:mod:`pyglet.media`
        does this for scheduling audio buffer updates, which need to occur
        regularly -- if all audio updates are scheduled at the same time
        (for example, mixing several tracks of a music score, or playing
        multiple videos back simultaneously), the resulting load on the
        CPU is excessive for those intervals but idle outside.  Using
        the soft interval scheduling, the load is more evenly distributed.

        Soft interval scheduling can also be used as an easy way to schedule
        graphics animations out of phase; for example, multiple flags
        waving in the wind.
        """
        next_ts = self._get_soft_next_ts(self._get_nearest_ts(), interval)
        last_ts = next_ts - interval
        item = _ScheduledIntervalItem(func, interval, last_ts, next_ts, args, kwargs)
        _heappush(self._schedule_interval_items, item)

    def unschedule(self, func: Callable) -> None:
        """Remove a function from the schedule.

        If the function appears in the schedule more than once, all occurrences
        are removed.  If the function was not scheduled, no error is raised.
        """
        # clever remove item without disturbing the heap:
        # 1. set function to an empty lambda -- original function is not called
        # 2. set interval to 0               -- item will be removed from heap eventually
        valid_items = {item for item in self._schedule_interval_items if item.func == func}

        if self._current_interval_item and self._current_interval_item.func == func:
            valid_items.add(self._current_interval_item)

        for item in valid_items:
            item.interval = 0
            item.func = lambda x, *args, **kwargs: x

        self._schedule_items = [i for i in self._schedule_items if i.func != func]


# Default clock.
_default = Clock()


def set_default(default: Clock) -> None:
    """Set the default clock to use for all module-level functions.

    By default, an instance of :py:class:`~pyglet.clock.Clock` is used.
    """
    global _default
    _default = default


def get_default() -> Clock:
    """Get the pyglet default Clock.

    Return the :py:class:`~pyglet.clock.Clock` instance that is used by all
    module-level clock functions.
    """
    return _default


def tick(poll: bool = False) -> float:
    """:see: :py:meth:`~pyglet.clock.Clock.tick`."""
    return _default.tick(poll)


def get_sleep_time(sleep_idle: bool) -> float | None:
    """:see: :py:meth:`~pyglet.clock.Clock.get_sleep_time`."""
    return _default.get_sleep_time(sleep_idle)


def get_frequency() -> float:
    """:see: :py:meth:`~pyglet.clock.Clock.get_frequency`."""
    return _default.get_frequency()


def schedule(func: Callable, *args: Any, **kwargs: Any) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.schedule`."""
    _default.schedule(func, *args, **kwargs)


def schedule_interval(func: Callable, interval: float, *args: Any, **kwargs: Any) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.schedule_interval`."""
    _default.schedule_interval(func, interval, *args, **kwargs)


def schedule_interval_for_duration(func: Callable, interval: float, duration: float, *args, **kwargs) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.schedule_interval_for_duration`."""
    _default.schedule_interval_for_duration(func, interval, duration, *args, **kwargs)


def schedule_interval_soft(func: Callable, interval: float, *args, **kwargs) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.schedule_interval_soft`."""
    _default.schedule_interval_soft(func, interval, *args, **kwargs)


def schedule_once(func: Callable, delay: float, *args, **kwargs) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.schedule_once`."""
    _default.schedule_once(func, delay, *args, **kwargs)


def unschedule(func: Callable) -> None:
    """:see: :py:meth:`~pyglet.clock.Clock.unschedule`."""
    _default.unschedule(func)
